#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Flow splitter depending on a Helmholtz EOS property package.  This just
multiplies the flow by the split fractions for the outlets and has a constraint
(sum_split) that requires the split fractions to sum to 1.  The usual way to
specify this unit is to have fix or have constraints that set the inlet, and
have n_outlets - 1 specified split fractions or outlet flows.

This model is pseudo-steady-state when used in dynamic mode.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Var, value
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale


__author__ = "John Eslick"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("HelmSplitter")
class HelmSplitterData(UnitModelBlockData):
    """
    This is a basic stream splitter which splits flow into outlet streams based
    on split fractions. This does not do phase separation, and assumes that you
    are using a Helmholtz EOS property package with P-H state variables. In
    dynamic mode this uses a pseudo-steady-state model.

    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for mixer",
            doc="""Property parameter object used to define property
calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_list",
        ConfigValue(
            domain=ListOf(str),
            description="List of outlet names",
            doc="""A list containing names of outlets,
**default** - None.
**Valid values:** {
**None** - use num_outlets argument,
**list** - a list of names to use for outlets.}""",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            domain=int,
            description="Number of outlets to unit",
            doc="""Argument indicating number (int) of outlets to construct,
not used if outlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use outlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of outlets to create (will be named with sequential integers
from 1 to num_outlets).}""",
        ),
    )

    def build(self):
        """
        Build a splitter.

        Args:
            None

        Returns:
            None
        """
        time = self.flowsheet().time
        super().build()

        self._get_property_package()

        self.create_outlet_list()
        self.add_inlet_state_and_port()
        self.add_outlet_state_blocks()
        self.add_outlet_port_objects()

        self.split_fraction = Var(
            time,
            self.outlet_list,
            initialize=1.0 / len(self.outlet_list),
            doc="Split fractions for outlet streams",
        )

        @self.Constraint(time, doc="Splt constraint")
        def sum_split(b, t):
            return 1 == sum(self.split_fraction[t, o] for o in self.outlet_list)

        @self.Constraint(time, self.outlet_list, doc="Pressure constraint")
        def pressure_eqn(b, t, o):
            o_block = getattr(self, "{}_state".format(o))
            return self.mixed_state[t].pressure == o_block[t].pressure

        @self.Constraint(time, self.outlet_list, doc="Enthalpy constraint")
        def enthalpy_eqn(b, t, o):
            o_block = getattr(self, "{}_state".format(o))
            return self.mixed_state[t].enth_mol == o_block[t].enth_mol

        @self.Constraint(time, self.outlet_list, doc="Flow constraint")
        def flow_eqn(b, t, o):
            o_block = getattr(self, "{}_state".format(o))
            sf = self.split_fraction[t, o]
            return self.mixed_state[t].flow_mol * sf == o_block[t].flow_mol

    def add_inlet_state_and_port(self):
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["defined_state"] = True
        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties of mixed (inlet) stream",
            **tmp_dict
        )
        self.add_port(name="inlet", block=self.mixed_state, doc="Inlet Port")

    def create_outlet_list(self):
        """
        Create list of outlet stream names based on config arguments.

        Returns:
            list of strings
        """
        config = self.config
        if config.outlet_list is not None and config.num_outlets is not None:
            # If both arguments provided and not consistent, raise Exception
            if len(config.outlet_list) != config.num_outlets:
                raise ConfigurationError(
                    "{} Splitter provided with both outlet_list and "
                    "num_outlets arguments, which were not consistent ("
                    "length of outlet_list was not equal to num_outlets). "
                    "Please check your arguments for consistency, and "
                    "note that it is only necessry to provide one of "
                    "these arguments.".format(self.name)
                )
        elif config.outlet_list is None and config.num_outlets is None:
            # If no arguments provided for outlets, default to num_outlets = 2
            config.num_outlets = 2

        # Create a list of names for outlet StateBlocks
        if config.outlet_list is not None:
            outlet_list = self.config.outlet_list
        else:
            outlet_list = [
                "outlet_{}".format(n) for n in range(1, config.num_outlets + 1)
            ]
        self.outlet_list = outlet_list

    def add_outlet_state_blocks(self):
        """
        Construct StateBlocks for all outlet streams.

        Args:
            None

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = False

        # Create empty list to hold StateBlocks for return
        self.outlet_blocks = {}

        # Create an instance of StateBlock for all outlets
        for o in self.outlet_list:
            o_obj = self.config.property_package.build_state_block(
                self.flowsheet().time, doc="Material properties at outlet", **tmp_dict
            )
            setattr(self, o + "_state", o_obj)
            self.outlet_blocks[o] = o_obj

    def add_outlet_port_objects(self):
        """
        Adds outlet Port objects if required.

        Args:
            None

        Returns:
            None
        """
        self.outlet_ports = {}
        for p in self.outlet_list:
            self.add_port(name=p, block=self.outlet_blocks[p], doc="Outlet")
            self.outlet_ports[p] = getattr(self, p)

    def initialize_build(self, outlvl=idaeslog.NOTSET, optarg=None, solver=None):
        """
        Initialization routine for splitter

        Keyword Arguments:
            outlvl: sets output level of initialization routine
            optarg: solver options dictionary object (default=None, use
                    default solver options)
            solver: str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # sp is what to save to make sure state after init is same as the start
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # check for fixed outlet flows and use them to calculate fixed split
        # fractions
        for t in self.flowsheet().time:
            for o in self.outlet_list:
                if self.outlet_blocks[o][t].flow_mol.fixed:
                    self.split_fraction[t, o].fix(
                        value(
                            self.mixed_state[t].flow_mol
                            / self.outlet_blocks[o][t].flow_mol
                        )
                    )

        # fix or unfix split fractions so n - 1 are fixed
        for t in self.flowsheet().time:
            # see how many split fractions are fixed
            n = sum(1 for o in self.outlet_list if self.split_fraction[t, o].fixed)
            # if number of outlets - 1 we're good
            if n == len(self.outlet_list) - 1:
                continue
            # if too mant are fixed un fix the first, generally assume that is
            # the main flow, and is the calculated split fraction
            if n == len(self.outlet_list):
                self.split_fraction[t, self.outlet_list[0]].unfix()
            # if not enough fixed, start fixing from the back until there are
            # are enough
            for o in reversed(self.outlet_list):
                if not self.split_fraction[t, o].fixed:
                    self.split_fraction[t, o].fix()
                    n += 1
                if n == len(self.outlet_list) - 1:
                    break

        # This model is really simple so it should easily solve without much
        # effort to initialize
        self.inlet.fix()
        for o, p in self.outlet_ports.items():
            p.unfix()
        assert degrees_of_freedom(self) == 0
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        from_json(self, sd=istate, wts=sp)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for (t, i), c in self.pressure_eqn.items():
            o_block = getattr(self, "{}_state".format(i))
            s = iscale.get_scaling_factor(o_block[t].pressure)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for (t, i), c in self.enthalpy_eqn.items():
            o_block = getattr(self, "{}_state".format(i))
            s = iscale.get_scaling_factor(o_block[t].enth_mol)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for (t, i), c in self.flow_eqn.items():
            o_block = getattr(self, "{}_state".format(i))
            s = iscale.get_scaling_factor(o_block[t].flow_mol)
            iscale.constraint_scaling_transform(c, s, overwrite=False)

    def _get_stream_table_contents(self, time_point=0):
        io_dict = {"inlet": self.inlet}
        for i in self.outlet_list:
            io_dict[i] = getattr(self, i)
        return create_stream_table_dataframe(io_dict, time_point=time_point)



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions for dynamic IDAES models.
"""

from pyomo.environ import Block, Constraint, Var
from pyomo.dae import DerivativeVar
from pyomo.dae.set_utils import (
    is_explicitly_indexed_by,
    is_in_block_indexed_by,
    get_index_set_except,
)
from pyomo.common.collections import ComponentSet

import idaes.logger as idaeslog

__author__ = "Robert Parker"


def get_activity_dict(b):
    """
    Function that builds a dictionary telling whether or not each
    ConstraintData and BlockData object in a model is active.
    Uses the objects' ids as the hash.

    Args:
        b : A Pyomo Block to be searched for active components

    Returns:
        A dictionary mapping id of constraint and block data objects
        to a bool indicating if they are active
    """
    # Note: active constraints/blocks contained in an inactive block will still
    # be marked as active.
    return {
        id(con): con.active for con in b.component_data_objects((Constraint, Block))
    }


def get_fixed_dict(b):
    """
    Function that builds a dictionary telling whether or not each VarData
    object in a model is fixed. Uses the objects' ids as the hash.

    Args:
        b : A Pyomo block to be searched for fixed variables

    Returns:
        A dictionary mapping id of VarData objects to a bool indicating if
        they are fixed
    """
    return {id(var): var.fixed for var in b.component_data_objects(Var)}


def deactivate_model_at(b, cset, pts, outlvl=idaeslog.NOTSET):
    """
    Finds any block or constraint in block b, indexed explicitly (and not
    implicitly) by cset, and deactivates it at points specified.
    Implicitly indexed components are excluded because one of their parent
    blocks will be deactivated, so deactivating them too would be redundant.

    Args:
        b : Block to search
        cset : ContinuousSet of interest
        pts : Value or list of values, in ContinuousSet, to deactivate at

    Returns:
        A dictionary mapping points in pts to lists of
        component data that have been deactivated there
    """
    if not isinstance(pts, list):
        pts = [pts]
    for pt in pts:
        if not pt in cset:
            msg = str(pt) + " is not in ContinuousSet " + cset.name
            raise ValueError(msg)
    deactivated = {pt: [] for pt in pts}

    visited = set()
    for comp in b.component_objects([Block, Constraint], active=True):
        # Record components that have been visited in case component_objects
        # contains duplicates (due to references)
        if id(comp) in visited:
            continue
        visited.add(id(comp))

        if is_explicitly_indexed_by(comp, cset) and not is_in_block_indexed_by(
            comp, cset
        ):
            info = get_index_set_except(comp, cset)
            non_cset_set = info["set_except"]
            index_getter = info["index_getter"]

            for non_cset_index in non_cset_set:
                for pt in pts:
                    index = index_getter(non_cset_index, pt)
                    try:
                        comp[index].deactivate()
                        deactivated[pt].append(comp[index])
                    except KeyError:
                        # except KeyError to allow Constraint/Block.Skip
                        msg = comp.name + " has no index " + str(index)
                        init_log = idaeslog.getInitLogger(__name__, outlvl)
                        init_log.warning(msg)
                        continue

    return deactivated


def deactivate_constraints_unindexed_by(b, time):
    """
    Searches block b for and constraints not indexed by time
    and deactivates them.

    Args:
        b : Block to search
        time : Set with respect to which to find unindexed constraints

    Returns:
        List of constraints deactivated
    """
    conlist = []

    visited = set()
    for comp in b.component_objects(Constraint, active=True):
        if id(comp) in visited:
            continue
        visited.add(id(comp))

        if not is_explicitly_indexed_by(comp, time) and not is_in_block_indexed_by(
            comp, time
        ):
            for index in comp:
                compdata = comp[index]
                if compdata.active:
                    compdata.deactivate()
                    conlist.append(compdata)

    return conlist


def fix_vars_unindexed_by(b, time):
    """
    Searches block b for variables not indexed by time
    and fixes them.

    Args:
        b : Block to search
        time : Set with respect to which to find unindexed variables

    Returns:
        List of variables fixed
    """
    varlist = []

    visited = set()
    for var in b.component_objects(Var):
        if id(var) in visited:
            continue
        visited.add(id(var))

        if not is_explicitly_indexed_by(var, time) and not is_in_block_indexed_by(
            var, time
        ):
            for index in var:
                vardata = var[index]
                if not vardata.fixed and vardata.value is not None:
                    # Can't fix a variable with a value of None, but this
                    # should be called after a solve, so any variable with
                    # value of None is stale and won't be sent to solver,
                    # so it doesn't need to be fixed to maintain correct
                    # degrees of freedom
                    vardata.fix()
                    varlist.append(vardata)

    return varlist


def get_location_of_coordinate_set(setprod, subset):
    """For a SetProduct and some 1-dimensional coordinate set of that
    SetProduct, returns the location of an index of the coordinate
    set within the index of the setproduct.

    Args:
        setprod : SetProduct containing the subset of interest
        subset : 1-dimensional set whose location will be found in the
                 SetProduct

    Returns:
        Integer location of the subset within the SetProduct
    """
    if subset.dimen != 1:
        # This could be supported in the future if there is demand for it
        raise ValueError(
            "Cannot get the location of %s because it is multi-dimensional"
            % (subset.name)
        )

    loc = None
    i = 0
    found = False
    if hasattr(setprod, "subsets"):
        subsets = setprod.subsets()
    elif hasattr(setprod, "set_tuple"):
        subsets = setprod.set_tuple
    else:
        subsets = [setprod]
    for _set in subsets:
        if _set is subset:
            if found:
                raise ValueError(
                    "Cannot get the location of %s because it appears "
                    "multiple times" % (_set.name)
                )
            found = True
            loc = i
            i += 1
        else:
            i += _set.dimen
    return loc


def get_index_of_set(comp, wrt):
    """For some data object of an indexed component, gets the value of the
    index corresponding to some 1-dimensional pyomo set.

    Args:
        comp : Component data object whose index will be searched
        wrt : Set whose index will be searched for

    Returns:
        Value of the specified set in the component data object
    """
    parent = comp.parent_component()
    if not is_explicitly_indexed_by(parent, wrt):
        raise ValueError(
            "Component %s is not explicitly indexed by set %s." % (comp.name, wrt.name)
        )

    index = comp.index()
    if not isinstance(index, tuple):
        index = (index,)
    loc = get_location_of_coordinate_set(parent.index_set(), wrt)
    return index[loc]


def get_implicit_index_of_set(comp, wrt):
    """For some data object contained (at some level of the hierarchy) in a
    block indexed by wrt, returns the index corresponding to wrt in that
    block.

    Args:
        comp : Component data object whose (parent blocks') indices will be
               searched
        wrt : Set whose index will be searched for

    Returns:
        Value of the specified set
    """
    val = None
    found = False
    if is_explicitly_indexed_by(comp.parent_component(), wrt):
        val = get_index_of_set(comp, wrt)
        found = True

    parent_block = comp.parent_block()
    while parent_block is not None:
        parent_component = parent_block.parent_component()
        if is_explicitly_indexed_by(parent_component, wrt):
            if found:
                raise ValueError(
                    "Cannot get the index of set %s because it appears "
                    "multiple times in the hierarchy" % (wrt.name)
                )
            val = get_index_of_set(parent_block, wrt)
            found = True
        parent_block = parent_block.parent_block()

    # Will return val even if it is None.
    # User can decide what to do in this case.
    return val


def get_derivatives_at(b, time, pts):
    """
    Finds derivatives with respect to time at points specified.
    No distinction made for multiple derivatives or mixed partials.

    Args:
        b : Block to search for derivatives
        time : ContinuousSet to look for derivatives with respect to
        pts : Value or list of values in time set at which to return
              derivatives

    Returns
        Dictionary mapping time points to lists of derivatives
        at those points
    """
    if not isinstance(pts, list):
        pts = [pts]
    dvdict = {pt: [] for pt in pts}

    visited = set()
    for var in b.component_objects(Var):
        if id(var) in visited:
            continue
        visited.add(id(var))

        if not isinstance(var, DerivativeVar):
            continue
        if time not in ComponentSet(var.get_continuousset_list()):
            continue

        info = get_index_set_except(var, time)
        non_time_set = info["set_except"]
        index_getter = info["index_getter"]
        for pt in pts:
            for non_time_index in non_time_set:
                index = index_getter(non_time_index, pt)
                dvdict[pt].append(var[index])

    return dvdict


# TODO: should be able to replace this function everywhere
#       with getname and find_component
#       ^ not true. Cannot call find_component from a BlockData object
#                   Or on a name containing a decimal index
#       component looks like a similar substitute for BlockDatas, but
#       cannot seem to call on names including indices at all
def path_from_block(comp, blk, include_comp=False):
    """
    Returns a list of tuples with (local_name, index) pairs required
    to locate comp from blk

    Args:
        comp : Component(Data) object to locate
        blk : Block(Data) to locate comp from
        include_comp : Bool of whether or not to include the
                       local_name, index of the component itself

    Returns:
        A list of string, index tuples that can be used to locate
        comp from blk
    """
    parent_data = comp.parent_block()
    route = []
    # Walk up the hierarchy tree until blk is reached
    while parent_data != blk:
        parent_obj = parent_data.parent_component()
        # Record the local name and index required to locate the current block
        # from the parent_component of its parent_block
        # Pre-pend to the existing (name, index) list
        route = [(parent_obj.local_name, parent_data.index())] + route
        # If top-levelmodel has been reached, break
        # (This should not happen)
        if parent_data.parent_block() is None:
            break
        parent_data = parent_data.parent_block()

    # Append comp's name and index to the list if desired:
    if include_comp:
        if hasattr(comp, "index"):
            route.append((comp.parent_component().local_name, comp.index()))
        else:
            # Not obvious what the right thing to do is if comp has no index
            # attribute...
            route.append((comp.parent_component().local_name, None))
    return route


def find_comp_in_block(tgt_block, src_block, src_comp, allow_miss=False):
    """This function finds a component in a source block, then uses the same
    local names and indices to try to find a corresponding component in a target
    block. This is used when we would like to verify that a component of the
    same name exists in the target block, as in model predictive control where
    certain variables must be correllated between plant and controller model.

    Args:
        tgt_block : Target block that will be searched for component
        src_block : Source block in which the original component is located
        src_comp : Component whose name will be searched for in target block
        allow_miss : If True, will ignore attribute and key errors due to
                     searching for non-existent components in the target model

    Returns:
        Component with the same name in the target block
    """

    local_parent = tgt_block
    for r in path_from_block(src_comp, src_block, include_comp=False):
        # Don't include comp as I want to use this to find IndexedComponents,
        # for which [r[1]] will result in a KeyError.
        try:
            local_parent = getattr(local_parent, r[0])[r[1]]
        except AttributeError:
            if allow_miss:
                return None
            else:
                raise AttributeError(
                    "%s has no attribute %s. Use allow_miss=True if this "
                    "is expected and acceptable." % (local_parent.name, r[0])
                )
        except KeyError:
            if allow_miss:
                return None
            else:
                raise KeyError(
                    "%s is not a valid index for %s, use allow_miss=True "
                    "if this is expected and acceptable."
                    % (str(r[1]), getattr(local_parent, r[0]).name)
                )

    # This logic should return the IndexedComponent or ComponentData,
    # whichever is appropriate
    try:
        tgt_comp = getattr(local_parent, src_comp.parent_component().local_name)
    except AttributeError:
        if allow_miss:
            return None
        else:
            raise AttributeError(
                "%s has no attribute %s. Use allow_miss=True if this "
                "is expected and acceptable."
                % (local_parent.name, src_comp.parent_component().local_name)
            )
    # tgt_comp is now an indexed component or simple component

    if hasattr(src_comp, "index"):
        # If comp has index, attempt to access it in tgt_comp
        index = src_comp.index()
        try:
            tgt_comp = tgt_comp[index]
        except KeyError:
            if allow_miss:
                return None
            else:
                raise KeyError(
                    "%s is not a valid index for %s, use allow_miss=True "
                    "if this is expected and acceptable." % (str(index), tgt_comp.name)
                )

    return tgt_comp


def find_comp_in_block_at_time(
    tgt_block, src_block, src_comp, time, t0, allow_miss=False
):
    """This function finds a component in a source block, then uses the same
    local names and indices to try to find a corresponding component in a target
    block, with the exception of time index in the target component, which is
    replaced by a specified time point. This is used for validation of a
    component by its name in the case where blocks may differ by at most time
    indices, for example validating a steady-state model or a model with a
    different time discretization.

    Args:
        tgt_block : Target block that will be searched for component
        src_block : Source block in which the original component is located
        src_comp : Component whose name will be searched for in target block
        time : Set whose index will be replaced in the target component
        t0 : Index of the time set that will be used in the target
             component
        allow_miss : If True, will ignore attribute and key errors due to
                     searching for non-existent components in the target model

    """
    # Could extend this to allow replacing indices of multiple sets
    # (useful for PDEs)

    if t0 not in time:
        raise KeyError("t0 must be in the time set")
    if time.model() is not tgt_block.model():
        raise ValueError("time must belong to the same model as the target block")
    if src_block.model() is not src_comp.model():
        raise ValueError("src_block and src_comp must be components of the same model")

    local_parent = tgt_block
    for r in path_from_block(src_comp, src_block, include_comp=False):
        # Don't include comp as I want to use this to find IndexedComponents,
        # for which [r[1]] will result in a KeyError.

        # If local_parent is indexed by time, need to replace time index
        # in r[1]

        try:
            local_parent = getattr(local_parent, r[0])
        except AttributeError:
            if allow_miss:
                return None
            else:
                raise AttributeError(
                    "%s has no attribute %s. Use allow_miss=True if this "
                    "is expected and acceptable." % (local_parent.name, r[0])
                )

        index = r[1]

        # Can abstract the following into a function:
        # replace_time_index or something
        if is_explicitly_indexed_by(local_parent, time):
            index_set = local_parent.index_set()
            time_loc = get_location_of_coordinate_set(index_set, time)

            if not isinstance(index, tuple):
                index = (index,)
            index = list(index)

            # Replace time index with t0
            index[time_loc] = t0
            index = tuple(index)

        try:
            local_parent = local_parent[index]
        except KeyError:
            if allow_miss:
                return None
            else:
                raise KeyError(
                    "%s is not a valid index for %s, use allow_miss=True "
                    "if this is expected and acceptable."
                    % (str(index), local_parent.name)
                )

    # This logic should return the IndexedComponent or ComponentData,
    # whichever is appropriate
    try:
        tgt_comp = getattr(local_parent, src_comp.parent_component().local_name)
    except AttributeError:
        if allow_miss:
            return None
        else:
            raise AttributeError(
                "%s has no attribute %s. Use allow_miss=True if this "
                "is expected and acceptable."
                % (local_parent.name, src_comp.parent_component().local_name)
            )
    # tgt_comp is now an indexed component or simple component

    if hasattr(src_comp, "index"):
        # If comp has index, attempt to access it in tgt_comp
        index = src_comp.index()

        if is_explicitly_indexed_by(tgt_comp, time):
            index_set = tgt_comp.index_set()
            time_loc = get_location_of_coordinate_set(index_set, time)

            if not isinstance(index, tuple):
                index = (index,)
            index = list(index)

            # Replace time index with t0
            index[time_loc] = t0
            index = tuple(index)

        try:
            tgt_comp = tgt_comp[index]
        except KeyError:
            if allow_miss:
                return None
            else:
                raise KeyError(
                    "%s is not a valid index for %s, use allow_miss=True "
                    "if this is expected and acceptable." % (str(index), tgt_comp.name)
                )

    return tgt_comp


def copy_non_time_indexed_values(
    fs_tgt,
    fs_src,
    copy_fixed=True,
    outlvl=idaeslog.NOTSET,
):
    """
    Function to set the values of all variables that are not (implicitly
    or explicitly) indexed by time to their values in a different flowsheet.

    Args:
        fs_tgt : Flowsheet into which values will be copied.
        fs_src : Flowsheet from which values will be copied.
        copy_fixed : Bool marking whether or not to copy over fixed variables
                     in the target flowsheet.
        outlvl : Outlevel for the IDAES logger.

    Returns:
        None
    """
    time_tgt = fs_tgt.time

    var_visited = set()
    for var_tgt in fs_tgt.component_objects(Var, descend_into=False):
        if id(var_tgt) in var_visited:
            continue
        var_visited.add(id(var_tgt))

        if is_explicitly_indexed_by(var_tgt, time_tgt):
            continue
        var_src = fs_src.find_component(var_tgt.local_name)
        # ^ this find_component is fine because var_tgt is a Var not VarData
        # and its local_name is used. Assumes that there are no other decimal
        # indices in between fs_src and var_src

        if var_src is None:
            # Log a warning
            msg = (
                "Warning copying values: "
                + var_src.name
                + " does not exist in source block "
                + fs_src.name
            )
            init_log = idaeslog.getInitLogger(__name__, outlvl)
            init_log.warning(msg)
            continue

        for index in var_tgt:
            if not copy_fixed and var_tgt[index].fixed:
                continue
            var_tgt[index].set_value(var_src.value)

    blk_visited = set()
    for blk_tgt in fs_tgt.component_objects(Block):

        if id(blk_tgt) in blk_visited:
            continue
        blk_visited.add(id(blk_tgt))

        if is_in_block_indexed_by(blk_tgt, time_tgt) or is_explicitly_indexed_by(
            blk_tgt, time_tgt
        ):
            continue
        # block is not even implicitly indexed by time
        for b_index in blk_tgt:

            var_visited = set()
            for var_tgt in blk_tgt[b_index].component_objects(Var, descend_into=False):
                if id(var_tgt) in var_visited:
                    continue
                var_visited.add(id(var_tgt))

                if is_explicitly_indexed_by(var_tgt, time_tgt):
                    continue

                # can't used find_component(local_name) here because I might
                # have decimal indices
                try:
                    local_parent = fs_src
                    for r in path_from_block(var_tgt, fs_tgt):
                        local_parent = getattr(local_parent, r[0])[r[1]]
                except AttributeError:
                    # log warning
                    msg = (
                        "Warning copying values: "
                        + r[0]
                        + " does not exist in source"
                        + local_parent.name
                    )
                    init_log = idaeslog.getInitLogger(__name__, outlvl)
                    init_log.warning(msg)
                    continue
                except KeyError:
                    msg = (
                        "Warning copying values: "
                        + str(r[1])
                        + " is not a valid index for"
                        + getattr(local_parent, r[0]).name
                    )
                    init_log = idaeslog.getInitLogger(__name__, outlvl)
                    init_log.warning(msg)
                    continue

                var_src = getattr(local_parent, var_tgt.local_name)

                for index in var_tgt:
                    if not copy_fixed and var_tgt[index].fixed:
                        continue
                    var_tgt[index].set_value(var_src[index].value)


def copy_values_at_time(
    fs_tgt, fs_src, t_target, t_source, copy_fixed=True, outlvl=idaeslog.NOTSET
):
    """
    Function to set the values of all (explicitly or implicitly) time-indexed
    variables in a flowsheet to similar values (with the same name) but at
    different points in time and (potentially) in different flowsheets.

    Args:
        fs_tgt : Target flowsheet, whose variables' values will get set
        fs_src : Source flowsheet, whose variables' values will be used to
                 set those of the target flowsheet. Could be the target
                 flowsheet
        t_target : Target time point
        t_source : Source time point
        copy_fixed : Bool of whether or not to copy over fixed variables in
                     target model
        outlvl : IDAES logger output level

    Returns:
        None
    """
    time_target = fs_tgt.time
    var_visited = set()
    for var_target in fs_tgt.component_objects(Var):
        if id(var_target) in var_visited:
            continue
        var_visited.add(id(var_target))

        if not is_explicitly_indexed_by(var_target, time_target):
            continue
        n = var_target.index_set().dimen

        local_parent = fs_src

        varname = var_target.getname(fully_qualified=True, relative_to=fs_tgt)
        # Calling find_component here makes the assumption that varname does not
        # contain decimal indices.
        var_source = fs_src.find_component(varname)
        if var_source is None:
            # Log a warning
            msg = (
                "Warning copying values: "
                + varname
                + " does not exist in source block "
                + fs_src.name
            )
            init_log = idaeslog.getInitLogger(__name__, outlvl)
            init_log.warning(msg)
            continue

        if n == 1:
            if not copy_fixed and var_target[t_target].fixed:
                continue
            var_target[t_target].set_value(var_source[t_source].value)
        elif n >= 2:
            index_info = get_index_set_except(var_target, time_target)
            non_time_index_set = index_info["set_except"]
            index_getter = index_info["index_getter"]
            for non_time_index in non_time_index_set:
                source_index = index_getter(non_time_index, t_source)
                target_index = index_getter(non_time_index, t_target)
                if not copy_fixed and var_target[target_index].fixed:
                    continue
                var_target[target_index].set_value(var_source[source_index].value)

    blk_visited = set()
    for blk_target in fs_tgt.component_objects(Block):
        if id(blk_target) in blk_visited:
            continue
        blk_visited.add(id(blk_target))

        if not is_explicitly_indexed_by(blk_target, time_target):
            continue
        n = blk_target.index_set().dimen

        blkname = blk_target.getname(fully_qualified=True, relative_to=fs_tgt)
        blk_source = fs_src.find_component(blkname)
        if blk_source is None:
            # log warning
            msg = (
                "Warning copying values: "
                + blkname
                + " does not exist in source"
                + fs_src.name
            )
            init_log = idaeslog.getInitLogger(__name__, outlvl)
            init_log.warning(msg)
            continue

        if n == 1:
            target_index = t_target
            source_index = t_source

            var_visited = set()
            for var_target in blk_target[target_index].component_data_objects(Var):
                if id(var_target) in var_visited:
                    continue
                var_visited.add(id(var_target))

                if not copy_fixed and var_target.fixed:
                    continue

                # Here, find_component will not work from BlockData object
                local_parent = blk_source[source_index]
                for r in path_from_block(var_target, blk_target[target_index]):
                    local_parent = getattr(local_parent, r[0])[r[1]]
                var_source = getattr(
                    local_parent, var_target.parent_component().local_name
                )[var_target.index()]
                var_target.set_value(var_source.value)

        elif n >= 2:
            index_info = get_index_set_except(blk_target, time_target)
            non_time_index_set = index_info["set_except"]
            index_getter = index_info["index_getter"]
            for non_time_index in non_time_index_set:
                source_index = index_getter(non_time_index, t_source)
                target_index = index_getter(non_time_index, t_target)

                var_visited = set()
                for var_target in blk_target[target_index].component_data_objects(Var):
                    if id(var_target) in var_visited:
                        continue
                    var_visited.add(id(var_target))

                    if not copy_fixed and var_target.fixed:
                        continue

                    local_parent = blk_source[source_index]
                    for r in path_from_block(var_target, blk_target[target_index]):
                        local_parent = getattr(local_parent, r[0])[r[1]]
                    var_source = getattr(
                        local_parent, var_target.parent_component().local_name
                    )[var_target.index()]
                    var_target.set_value(var_source.value)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
##############################################################################
# The development of this flowsheet/code is funded by the ARPA-E DIFFERENTIATE
# project: “Machine Learning for Natural Gas to Electric Power System Design”
# Project number: DE-FOA-0002107-1625.
# This project is a collaborative effort between the Pacific Northwest National
# Laboratory, the National Energy Technology Laboratory, and the University of
# Washington to design NGFC systems with high efficiencies and low CO2
# emissions.
##############################################################################
"""
This is a surrogate (black box) model for a CO2 purification process
based on cryogenic distillation.

The unit has 1 inlet stream for the CO2-rich flue gas: inlet
The unit has 3 outlet streams: pureco2, water, and vent

The state variables are flow_mol, mole_frac_comp, temperature, and pressure.
The state variables can be accessed via ports named: inlet, pureco2, water,
and vent

The surrogate models were prepared by varying the inlet flowrate and
composition. The expected value is 11350 lbmol/hr, lower bound: 10850 lbmol/hr,
and upper bound: 11850 lbmol/hr.
300 samples have been developed using Latin Hypercube Sampling
Method, the data set is trained by running the CPU rigorous model. ALAMO was
used to fit the surrogate models.
Outputs are a function of the inlet flowrate and species composition.
Inlet temperature and pressure are assumed to be fixed at 100 F and 14.7 psia.
The degrees of freedom are the inlet states:
(1) Inlet flow in mol/s
(2) Inlet component mole fractions: CO2, H2O, N2, Ar, O2
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Var, units as pyunits
from pyomo.network import Port

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver

__author__ = "Differentiate Team (N. Susarla, A. Noring, M. Zamarripa)"
__version__ = "1.1.0"


# ----------------------------------------------------------------------------
@declare_process_block_class("CarbonProcessingUnit")
class CarbonProcessingUnitData(UnitModelBlockData):
    """
    This surrogate model is based on flow of CO2 rich stream flow only
    Assumptions:
    Fixed temperature, and pressure of feed stream
    Fixed CO2 purity in the CO2 product stream
    """

    def build(self):
        self.component_list = ["Ar", "CO2", "O2", "H2O", "N2"]

        self._make_vars()
        self._add_material_balances()
        self._add_surrogates()

        # Add ports: 4 (1 for inlet and 3 for outlets)
        self.inlet = Port(noruleinit=True, doc="A port for the co2 rich inlet stream")
        self.pureco2 = Port(
            noruleinit=True, doc="A port for the pure CO2 outlet stream"
        )
        self.water = Port(noruleinit=True, doc="A port for the water outlet stream")
        self.vent = Port(noruleinit=True, doc="A port for thevent gas outlet stream")

        # Add state vars to the ports
        self.inlet.add(self.inlet_flow_mol, "flow_mol")
        self.inlet.add(self.inlet_temperature, "temperature")
        self.inlet.add(self.inlet_pressure, "pressure")
        self.inlet.add(self.inlet_mole_frac_comp, "mole_frac_comp")

        self.pureco2.add(self.pureco2_flow_mol, "flow_mol")
        self.pureco2.add(self.pureco2_temperature, "temperature")
        self.pureco2.add(self.pureco2_pressure, "pressure")
        self.pureco2.add(self.pureco2_mole_frac_comp, "mole_frac_comp")

        self.water.add(self.water_flow_mol, "flow_mol")
        self.water.add(self.water_temperature, "temperature")
        self.water.add(self.water_pressure, "pressure")
        self.water.add(self.water_mole_frac_comp, "mole_frac_comp")

        self.vent.add(self.vent_flow_mol, "flow_mol")
        self.vent.add(self.vent_temperature, "temperature")
        self.vent.add(self.vent_pressure, "pressure")
        self.vent.add(self.vent_mole_frac_comp, "mole_frac_comp")

    def _make_vars(self):
        """This section is for creating all the vars for this model.
        There are 1 inlet and 3 outlet streams.
        These streams are names as inlet, pureco2, water, vent
        For each of these streams, the following variables are defined:
        (1) Total mole flow [mol/s]: [stream_name]_flow_mol
        (2) Component molar fraction: [stream_name]_mole_frac_comp
        (3) Component mole flows [mol/s]: [stream_name]_flow_mol_comp
        (4) Temperature [K]: [stream_name]_temperature
        (5) Pressure [Pa]: [stream_name]_pressure
        """

        # units declaration for vars
        flow_units = pyunits.mol / pyunits.s
        pressure_units = pyunits.Pa
        temperature_units = pyunits.K
        work_units = pyunits.J / pyunits.s

        # Add vars for stream total mole flows
        self.inlet_flow_mol = Var(
            self.flowsheet().config.time, initialize=3600, units=flow_units
        )
        self.pureco2_flow_mol = Var(
            self.flowsheet().config.time, initialize=900, units=flow_units
        )
        self.water_flow_mol = Var(
            self.flowsheet().config.time, initialize=300, units=flow_units
        )
        self.vent_flow_mol = Var(
            self.flowsheet().config.time, initialize=2400, units=flow_units
        )

        # Add vars for stream component mole fractions
        self.inlet_mole_frac_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=1 / len(self.component_list),
        )
        self.pureco2_mole_frac_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=1 / len(self.component_list),
        )
        self.water_mole_frac_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=1 / len(self.component_list),
        )
        self.vent_mole_frac_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=1 / len(self.component_list),
        )

        # Add vars for stream component mole flows
        self.inlet_flow_mol_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=3600 / len(self.component_list),
            units=flow_units,
        )
        self.pureco2_flow_mol_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=900 / len(self.component_list),
            units=flow_units,
        )
        self.water_flow_mol_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=300 / len(self.component_list),
            units=flow_units,
        )
        self.vent_flow_mol_comp = Var(
            self.flowsheet().config.time,
            self.component_list,
            initialize=2400 / len(self.component_list),
            units=flow_units,
        )

        # Add vars for stream temperatures
        self.inlet_temperature = Var(
            self.flowsheet().config.time, initialize=110, units=temperature_units
        )
        self.pureco2_temperature = Var(
            self.flowsheet().config.time, initialize=110, units=temperature_units
        )
        self.water_temperature = Var(
            self.flowsheet().config.time, initialize=110, units=temperature_units
        )
        self.vent_temperature = Var(
            self.flowsheet().config.time, initialize=110, units=temperature_units
        )

        # Add vars for stream pressures
        self.inlet_pressure = Var(
            self.flowsheet().config.time, initialize=17, units=pressure_units
        )
        self.pureco2_pressure = Var(
            self.flowsheet().config.time, initialize=17, units=pressure_units
        )
        self.water_pressure = Var(
            self.flowsheet().config.time, initialize=17, units=pressure_units
        )
        self.vent_pressure = Var(
            self.flowsheet().config.time, initialize=17, units=pressure_units
        )

        self.heat_duty = Var(
            self.flowsheet().config.time,
            initialize=0,
            units=work_units,
            doc="CO2 compressor cooling duty",
        )

        self.work = Var(
            self.flowsheet().config.time,
            initialize=0,
            units=work_units,
            doc="CO2 compressor work",
        )

    def _add_material_balances(self):
        # Sum of all components mole fractions in a stream equals 1
        @self.Constraint(self.flowsheet().config.time)
        def mole_frac_comp_pureco2_eqn(b, t):
            return 0 == 1 - sum(
                b.pureco2_mole_frac_comp[t, c] for c in self.component_list
            )

        @self.Constraint(self.flowsheet().config.time)
        def mole_frac_comp_water_eqn(b, t):
            return 0 == 1 - sum(
                b.water_mole_frac_comp[t, c] for c in self.component_list
            )

        @self.Constraint(self.flowsheet().config.time)
        def mole_frac_comp_vent_eqn(b, t):
            return 0 == 1 - sum(
                b.vent_mole_frac_comp[t, c] for c in self.component_list
            )

        # Component mole flow = total mole flow * mole fracaction
        # There is one constraint for each stream
        @self.Constraint(self.flowsheet().config.time, self.component_list)
        def flow_mol_comp_inlet_eqn(b, t, c):
            return (
                b.inlet_flow_mol_comp[t, c]
                == b.inlet_flow_mol[t] * b.inlet_mole_frac_comp[t, c]
            )

        @self.Constraint(self.flowsheet().config.time, self.component_list)
        def flow_mol_comp_pureco2_eqn(b, t, c):
            return (
                b.pureco2_flow_mol_comp[t, c]
                == b.pureco2_flow_mol[t] * b.pureco2_mole_frac_comp[t, c]
            )

        @self.Constraint(self.flowsheet().config.time, self.component_list)
        def flow_mol_comp_water_eqn(b, t, c):
            return (
                b.water_flow_mol_comp[t, c]
                == b.water_flow_mol[t] * b.water_mole_frac_comp[t, c]
            )

        @self.Constraint(self.flowsheet().config.time, self.component_list)
        def flow_mol_comp_vent_eqn(b, t, c):
            return (
                b.vent_flow_mol_comp[t, c]
                == b.vent_flow_mol[t] * b.vent_mole_frac_comp[t, c]
            )

        # Mass balance for each component across the unit
        @self.Constraint(self.flowsheet().config.time, self.component_list)
        def component_material_balance_eqn(b, t, c):
            return (
                0
                == self.inlet_flow_mol_comp[t, c]
                - self.pureco2_flow_mol_comp[t, c]
                - self.water_flow_mol_comp[t, c]
                - self.vent_flow_mol_comp[t, c]
            )

    def _add_surrogates(self):
        # Compressor heat duty
        @self.Constraint(self.flowsheet().config.time)
        def heat_duty_eq(b, t):
            return b.heat_duty[t] == heat_duty_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Compressor train work
        @self.Constraint(self.flowsheet().config.time)
        def work_eq(b, t):
            return b.work[t] == compressor_power_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: Temperature
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_temperature_eq(b, t):
            return b.pureco2_temperature[t] == pureco2_temperature_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: Pressure
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_pressure_eq(b, t):
            return b.pureco2_pressure[t] == pureco2_pressure_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: Total mole flow
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_total_flow_eq(b, t):
            return b.pureco2_flow_mol[t] == pureco2_flow_mol_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: CO2 mole flow
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_co2_flow_mol_comp_eq(b, t):
            return b.pureco2_flow_mol_comp[t, "CO2"] == pureco2_co2_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: O2 mole flow
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_o2_flow_mol_comp_eq(b, t):
            return b.pureco2_flow_mol_comp[t, "O2"] == pureco2_o2_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: Ar mole flow
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_ar_flow_mol_comp_eq(b, t):
            return b.pureco2_flow_mol_comp[t, "Ar"] == pureco2_ar_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Pure CO2 stream: H2O mole flow
        @self.Constraint(self.flowsheet().config.time)
        def pureco2_h2o_flow_mol_comp_eq(b, t):
            return b.pureco2_flow_mol_comp[t, "H2O"] == pureco2_h2o_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: Total mole flow
        @self.Constraint(self.flowsheet().config.time)
        def water_total_flow_eq(b, t):
            return b.water_flow_mol[t] == water_flow_mol_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: Temperature
        @self.Constraint(self.flowsheet().config.time)
        def water_temperature_eq(b, t):
            return b.water_temperature[t] == water_temperature_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: Pressure
        @self.Constraint(self.flowsheet().config.time)
        def water_pressure_eq(b, t):
            return b.water_pressure[t] == water_pressure_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: CO2 mole flow
        @self.Constraint(self.flowsheet().config.time)
        def water_co2_flow_mol_comp_eq(b, t):
            return b.water_flow_mol_comp[t, "CO2"] == water_co2_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: O2 mole flow
        @self.Constraint(self.flowsheet().config.time)
        def water_o2_flow_mol_comp_eq(b, t):
            return b.water_flow_mol_comp[t, "O2"] == water_o2_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: Ar mole flow
        @self.Constraint(self.flowsheet().config.time)
        def water_ar_flow_mol_comp_eq(b, t):
            return b.water_flow_mol_comp[t, "Ar"] == water_ar_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Water stream: H2O mole flow
        @self.Constraint(self.flowsheet().config.time)
        def water_h2o_flow_mol_comp_eq(b, t):
            return b.water_flow_mol_comp[t, "H2O"] == water_h2o_flow_mol_comp_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Vent stream: Temperature
        @self.Constraint(self.flowsheet().config.time)
        def vent_temperature_eq(b, t):
            return b.vent_temperature[t] == vent_temperature_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

        # Vent stream: Pressure
        @self.Constraint(self.flowsheet().config.time)
        def vent_pressure_eq(b, t):
            return b.vent_pressure[t] == vent_pressure_fun(
                b.inlet_flow_mol[t],
                b.inlet_mole_frac_comp[t, "Ar"],
                b.inlet_mole_frac_comp[t, "CO2"],
                b.inlet_mole_frac_comp[t, "O2"],
                b.inlet_mole_frac_comp[t, "H2O"],
                b.inlet_mole_frac_comp[t, "N2"],
            )

    def initialize(
        blk,
        outlvl=idaeslog.NOTSET,
        solver="ipopt",
        optarg=None,
        release_state=True,
    ):
        """
        CO2 pure pyomo block initialization routine

        Keyword Arguments:
            outlvl : sets output level of initialization routine

            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating which solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        """
        if optarg is None:
            optarg = {"tol": 1e-6}

        iscale.calculate_scaling_factors(blk)
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        opt = get_solver(solver="ipopt", options=optarg)

        init_log.info_low("Starting initialization...")

        blk.inlet.flow_mol[0].fix()
        blk.inlet.mole_frac_comp[0, "Ar"].fix()
        blk.inlet.mole_frac_comp[0, "CO2"].fix()
        blk.inlet.mole_frac_comp[0, "O2"].fix()
        blk.inlet.mole_frac_comp[0, "H2O"].fix()
        blk.inlet.mole_frac_comp[0, "N2"].fix()

        # solve model
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 1 {}.".format(idaeslog.condition(res)))
        init_log.info_high("Initialization Step 1 Complete.")

        # release all states and activate constraint
        if release_state:
            blk.inlet.flow_mol[0].unfix()
            blk.inlet.mole_frac_comp[0, "Ar"].unfix()
            blk.inlet.mole_frac_comp[0, "CO2"].unfix()
            blk.inlet.mole_frac_comp[0, "O2"].unfix()
            blk.inlet.mole_frac_comp[0, "H2O"].unfix()
            blk.inlet.mole_frac_comp[0, "N2"].unfix()

        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for t in self.flowsheet().time:

            iscale.constraint_scaling_transform(self.mole_frac_comp_pureco2_eqn[t], 1)

            iscale.constraint_scaling_transform(self.mole_frac_comp_water_eqn[t], 1)

            iscale.constraint_scaling_transform(self.mole_frac_comp_vent_eqn[t], 1)

            iscale.constraint_scaling_transform(self.heat_duty_eq[t], 1e-6)

            iscale.constraint_scaling_transform(self.work_eq[t], 1e-6)

            iscale.constraint_scaling_transform(self.pureco2_temperature_eq[t], 1e-2)

            iscale.constraint_scaling_transform(self.pureco2_pressure_eq[t], 1e-5)

            iscale.constraint_scaling_transform(self.pureco2_total_flow_eq[t], 1e-2)

            iscale.constraint_scaling_transform(
                self.pureco2_co2_flow_mol_comp_eq[t], 1e-2
            )

            iscale.constraint_scaling_transform(self.pureco2_o2_flow_mol_comp_eq[t], 1)

            iscale.constraint_scaling_transform(self.pureco2_ar_flow_mol_comp_eq[t], 1)

            iscale.constraint_scaling_transform(
                self.pureco2_h2o_flow_mol_comp_eq[t], 1e-2
            )

            iscale.constraint_scaling_transform(self.water_total_flow_eq[t], 1e-2)

            iscale.constraint_scaling_transform(self.water_temperature_eq[t], 1e-2)

            iscale.constraint_scaling_transform(self.water_pressure_eq[t], 1e-5)

            iscale.constraint_scaling_transform(
                self.water_co2_flow_mol_comp_eq[t], 1e-2
            )

            iscale.constraint_scaling_transform(self.water_o2_flow_mol_comp_eq[t], 1)

            iscale.constraint_scaling_transform(self.water_ar_flow_mol_comp_eq[t], 1)

            iscale.constraint_scaling_transform(
                self.water_h2o_flow_mol_comp_eq[t], 1e-2
            )

            iscale.constraint_scaling_transform(self.vent_temperature_eq[t], 1e-2)

            iscale.constraint_scaling_transform(self.vent_pressure_eq[t], 1e-5)

            for c in self.component_list:

                iscale.constraint_scaling_transform(
                    self.flow_mol_comp_inlet_eqn[t, c], 1e-2
                )

                iscale.constraint_scaling_transform(
                    self.flow_mol_comp_pureco2_eqn[t, c], 1e-2
                )

                iscale.constraint_scaling_transform(
                    self.flow_mol_comp_water_eqn[t, c], 1e-2
                )

                iscale.constraint_scaling_transform(
                    self.flow_mol_comp_vent_eqn[t, c], 1e-2
                )

                iscale.constraint_scaling_transform(
                    self.component_material_balance_eqn[t, c], 1e-2
                )

        iscale.set_scaling_factor(self.inlet_flow_mol[0.0], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol[0.0], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol[0.0], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol[0.0], 1e-3)
        iscale.set_scaling_factor(self.inlet_flow_mol_comp[0.0, "CO2"], 1e-3)
        iscale.set_scaling_factor(self.inlet_flow_mol_comp[0.0, "H2O"], 1e-3)
        iscale.set_scaling_factor(self.inlet_flow_mol_comp[0.0, "O2"], 1e-3)
        iscale.set_scaling_factor(self.inlet_flow_mol_comp[0.0, "Ar"], 1e-3)
        iscale.set_scaling_factor(self.inlet_flow_mol_comp[0.0, "N2"], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol_comp[0.0, "CO2"], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol_comp[0.0, "H2O"], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol_comp[0.0, "O2"], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol_comp[0.0, "Ar"], 1e-3)
        iscale.set_scaling_factor(self.pureco2_flow_mol_comp[0.0, "N2"], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol_comp[0.0, "CO2"], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol_comp[0.0, "H2O"], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol_comp[0.0, "O2"], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol_comp[0.0, "Ar"], 1e-3)
        iscale.set_scaling_factor(self.water_flow_mol_comp[0.0, "N2"], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol_comp[0.0, "CO2"], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol_comp[0.0, "H2O"], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol_comp[0.0, "O2"], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol_comp[0.0, "Ar"], 1e-3)
        iscale.set_scaling_factor(self.vent_flow_mol_comp[0.0, "N2"], 1e-3)
        iscale.set_scaling_factor(self.inlet_temperature[0.0], 1e-2)
        iscale.set_scaling_factor(self.pureco2_temperature[0.0], 1e-2)
        iscale.set_scaling_factor(self.water_temperature[0.0], 1e-2)
        iscale.set_scaling_factor(self.vent_temperature[0.0], 1e-2)
        iscale.set_scaling_factor(self.inlet_pressure[0.0], 1e-5)
        iscale.set_scaling_factor(self.pureco2_pressure[0.0], 1e-5)
        iscale.set_scaling_factor(self.water_pressure[0.0], 1e-5)
        iscale.set_scaling_factor(self.vent_pressure[0.0], 1e-5)
        iscale.set_scaling_factor(self.heat_duty, 1e-6)
        iscale.set_scaling_factor(self.work, 1e-6)


def compressor_power_fun(x1, x2, x3, x4, x5, x6):
    return (
        16592.309504827182536246 * x1
        + 17806584.568199951201677 * x3
        - 15529374.443028066307306
    )


def heat_duty_fun(x1, x2, x3, x4, x5, x6):
    return (
        27431.344430845900205895 * x1
        + 25040167.356691692024469 * x3
        - 21837134.557537198066711
    )


def pureco2_flow_mol_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.84491492199656215156267 * x1
        + 1381.8486099781443954271 * x3
        - 1204.9518833417739642755
    )


def pureco2_ar_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.87539452852376016910194e-007 * x1
        + 0.31019041198388018487631e-001 * x2
        - 0.16384578403083416373726e-002 * x5
        - 0.11401484796994311183421e-002 * x6
    )


def pureco2_co2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.84491226339466729555738 * x1
        + 1381.8776912884593457420 * x3
        - 1204.9772340007211823831
    )


def pureco2_o2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.25557869821781261075201e-005 * x1
        - 0.30160307289038564698691e-001 * x3
        + 0.26298816057506283622169e-001
    )


def pureco2_h2o_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return 0.00


def pureco2_n2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return 0.69246160597454559832190e-006 * x3 - 0.56005883473002498644045e-006


def pureco2_temperature_fun(x1, x2, x3, x4, x5, x6):
    return (
        10.267585306350582641244 * x2
        + 32.023906991174357017371 * x3
        + 30.302544593098598824099 * x4
        + 278.86088998572887476257
    )


def pureco2_pressure_fun(x1, x2, x3, x4, x5, x6):
    return -0.28666413419197245189076e-011 * x1 + 15271893.400000000372529


def water_flow_mol_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.64898275330398047522351e-001 * x1
        + 995.23584715669267097837 * x4
        - 64.595577093201725915605
    )


def water_ar_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.23736607897806708417605e-008 * x1
        + 0.83689507888515507049582e-003 * x2
        - 0.14749984695782550662112e-003 * x5
    )


def water_co2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.36504473074529521385892e-005 * x1
        + 0.39821667908625070844697e-002 * x3
        + 0.30283921974618616818065e-001 * x4
        - 0.54375872162499300915828e-002
    )


def water_o2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.46405004009352215417855e-007 * x1
        - 0.99119387109367312720110e-003 * x3
        + 0.86473506719933338534462e-003
    )


def water_h2o_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.64894576227660510925332e-001 * x1
        + 995.21157061827375400753 * x4
        - 64.594007608678808196601
    )


def water_n2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return 0.71432912211586127618749e-009 * x1


def water_temperature_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.47345180660722969989695e-002 * x3
        - 1.1291949419088465056404 * x4
        - 0.13703395454789617582958e-001 * x5
        + 311.08988060641769379799
    )


def water_pressure_fun(x1, x2, x3, x4, x5, x6):
    return 117210.91999999999825377


def vent_flow_mol_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.90250925128294370858306e-001 * x1
        - 1373.8223230605847220431 * x3
        + 1197.7852419317096064333
    )


def vent_ar_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.27178896806122057258626e-002 * x1
        + 953.43145974385458885081 * x2
        - 168.06049873202732669597 * x5
    )


def vent_co2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.27083917235717330090905e-001 * x1
        - 407.85524106058562665567 * x3
        + 355.59132055263916072363
    )


def vent_o2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return (
        0.44846813052226472406936e-001 * x1
        - 956.47536271014735120843 * x3
        + 834.42603654916342748038
    )


def vent_h2o_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return 0.00


def vent_n2_flow_mol_comp_fun(x1, x2, x3, x4, x5, x6):
    return 0.15024528735008041077648e-001 * x1


def vent_temperature_fun(x1, x2, x3, x4, x5, x6):
    return -0.43741475554195015242121e-016 * x1 + 305.37222222222220580079


def vent_pressure_fun(x1, x2, x3, x4, x5, x6):
    return -0.22395635483747847803966e-013 * x1 + 115831.96800000000803266



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for unit_model.

Author: Andrew Lee
"""
import pytest
import re

from pyomo.environ import Block, ConcreteModel, Constraint, Var, units
from pyomo.network import Port

from idaes.core import (
    FlowsheetBlockData,
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    ControlVolume0DBlock,
    ControlVolume1DBlock,
    MaterialBalanceType,
    FlowDirection,
)
from idaes.core.util.exceptions import (
    BalanceTypeNotSupportedError,
    ConfigurationError,
    DynamicError,
)
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.initialization import SingleControlVolumeUnitInitializer


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class UnitData(UnitModelBlockData):
    def build(self):
        super(UnitModelBlockData, self).build()


@pytest.mark.unit
def test_config_block():
    m = ConcreteModel()

    m.u = Unit()

    assert len(m.u.config) == 2
    assert m.u.config.dynamic == useDefault

    assert m.u.default_initializer is SingleControlVolumeUnitInitializer


@pytest.mark.unit
def test_config_args():
    m = ConcreteModel()

    m.u = Unit(dynamic=True)

    assert m.u.config.dynamic is True


@pytest.mark.unit
def test_config_args_invalid():
    # Test validation of config arguments
    m = ConcreteModel()

    m.u = Unit()

    m.u.config.dynamic = True
    m.u.config.dynamic = False
    m.u.config.dynamic = None

    # Test that Value error raised when given invalid config arguments
    with pytest.raises(ValueError):
        m.u.config.dynamic = "foo"  # invalid str
    with pytest.raises(ValueError):
        m.u.config.dynamic = 5  # invalid int
    with pytest.raises(ValueError):
        m.u.config.dynamic = 2.0  # invalid float
    with pytest.raises(ValueError):
        m.u.config.dynamic = [2.0]  # invalid list
    with pytest.raises(ValueError):
        m.u.config.dynamic = {"a": 2.0}  # invalid dict


@pytest.mark.unit
def test_setup_dynamics1():
    # Test that _setup_dynamics gets argument from parent
    m = ConcreteModel()

    m.fs = Flowsheet(dynamic=False)

    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    assert m.fs.u.config.dynamic is False


@pytest.mark.unit
def test_setup_dynamics2():
    # Test that _setup_dynamics returns an DynamicError when parent has no
    # dynamic config argument

    m = ConcreteModel()
    # Intermediate Block is required, as ConcreteModels have a config attribute
    m.b = Block()
    m.b.u = Unit()

    with pytest.raises(DynamicError):
        m.b.u._setup_dynamics()


@pytest.mark.unit
def test_setup_dynamics_dynamic_in_steady_state():
    # Test that a DynamicError is raised when a dynamic models is placed in a
    # steady-state parent
    m = ConcreteModel()

    m.fs = Flowsheet(dynamic=False)

    m.fs.u = Unit(dynamic=True)
    with pytest.raises(DynamicError):
        m.fs.u._setup_dynamics()


@pytest.mark.unit
def test_setup_dynamics_get_time():
    # Test that time domain is collected correctly
    m = ConcreteModel()

    m.fs = Flowsheet(dynamic=False)

    m.fs.u = Unit()
    m.fs.u._setup_dynamics()


@pytest.mark.unit
def test_setup_dynamics_has_holdup():
    # Test that has_holdup argument is True when dynamic is True
    m = ConcreteModel()

    m.fs = Flowsheet(dynamic=True, time_units=units.s)

    m.fs.u = Unit()
    m.fs.u.config.has_holdup = False

    with pytest.raises(ConfigurationError):
        m.fs.u._setup_dynamics()


@pytest.mark.unit
def test_add_port():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.prop = m.fs.pp.build_state_block(m.fs.time)

    p_obj = m.fs.u.add_port(name="test_port", block=m.fs.u.prop)

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1
    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.test_port.component_flow_phase[0, p, j].value
                == m.fs.u.prop[0].flow_mol_phase_comp[p, j].value
            )
    assert m.fs.u.test_port.pressure[0].value == m.fs.u.prop[0].pressure.value
    assert m.fs.u.test_port.temperature[0].value == m.fs.u.prop[0].temperature.value


@pytest.mark.unit
def test_add_port_invalid_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.prop = m.fs.pp.build_state_block(m.fs.time)

    with pytest.raises(
        ConfigurationError,
        match=re.escape(
            "fs.u block object provided to add_port method is not an "
            "instance of a StateBlock object (does not have a build_port method)."
        ),
    ):
        m.fs.u.add_port(name="test_port", block=m.fs.u)


@pytest.mark.unit
def test_add_inlet_port_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.inlet is p_obj
    assert len(m.fs.u.inlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.inlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties_in[0].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.inlet.pressure[0] is m.fs.u.control_volume.properties_in[0].pressure
    assert (
        m.fs.u.inlet.temperature[0]
        is m.fs.u.control_volume.properties_in[0].temperature
    )


@pytest.mark.unit
def test_add_inlet_port_CV1D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume1DBlock(
        property_package=m.fs.pp,
        transformation_method="dae.finite_difference",
        transformation_scheme="BACKWARD",
    )

    m.fs.u.control_volume.add_geometry()
    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.inlet is p_obj
    assert len(m.fs.u.inlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.inlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties[0, 0].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.inlet.pressure[0] is m.fs.u.control_volume.properties[0, 0].pressure
    assert (
        m.fs.u.inlet.temperature[0]
        is m.fs.u.control_volume.properties[0, 0].temperature
    )


@pytest.mark.unit
def test_add_inlet_port_CV1D_backward():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume1DBlock(
        property_package=m.fs.pp,
        transformation_method="dae.finite_difference",
        transformation_scheme="BACKWARD",
    )

    m.fs.u.control_volume.add_geometry(flow_direction=FlowDirection.backward)
    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.inlet is p_obj
    assert len(m.fs.u.inlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.inlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties[0, 1].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.inlet.pressure[0] is m.fs.u.control_volume.properties[0, 1].pressure
    assert (
        m.fs.u.inlet.temperature[0]
        is m.fs.u.control_volume.properties[0, 1].temperature
    )


@pytest.mark.unit
def test_add_inlet_port_CV0D_no_default_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    with pytest.raises(
        ConfigurationError,
        match=re.escape(
            "fs.u add_inlet_port was called without a block argument"
            " but no default ControlVolume exists (control_volume). "
            "Please provide block to which the Port should be associated."
        ),
    ):
        m.fs.u.add_inlet_port()


@pytest.mark.unit
def test_add_inlet_port_CV0D_full_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port(name="test_port", block=m.fs.u.cv, doc="Test")

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1

    # Set new inlet conditions to differentiate from outlet
    m.fs.u.cv.properties_in[0].a = 10
    m.fs.u.cv.properties_in[0].b = 20
    m.fs.u.cv.properties_in[0].c = 30

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert m.fs.u.test_port.component_flow_phase[0, p, j].value == (
                m.fs.u.cv.properties_in[0].flow_mol_phase_comp[p, j].value
            )
    assert (
        m.fs.u.test_port.pressure[0].value == m.fs.u.cv.properties_in[0].pressure.value
    )
    assert (
        m.fs.u.test_port.temperature[0].value
        == m.fs.u.cv.properties_in[0].temperature.value
    )


@pytest.mark.unit
def test_add_inlet_port_CV0D_part_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(
        ConfigurationError,
        match="fs.u add_inlet_port was called with a block argument, "
        "but a name argument was not provided. Either both "
        "a name and a block must be provided or neither.",
    ):
        m.fs.u.add_inlet_port(block=m.fs.u.cv)

    with pytest.raises(
        ConfigurationError,
        match="fs.u add_inlet_port was called without a block "
        "argument but a name argument was provided. Either "
        "both a name and a block must be provided or neither.",
    ):
        m.fs.u.add_inlet_port(name="foo")


@pytest.mark.unit
def test_add_inlet_port_CV0D_no_state_blocks():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    with pytest.raises(
        ConfigurationError,
        match="fs.u - control volume does not have expected "
        "names for StateBlocks. Please check that the control "
        "volume was constructed correctly.",
    ):
        m.fs.u.add_inlet_port()


@pytest.mark.unit
def test_add_outlet_port_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.outlet is p_obj
    assert len(m.fs.u.outlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.outlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties_out[0].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.outlet.pressure[0] is m.fs.u.control_volume.properties_out[0].pressure
    assert (
        m.fs.u.outlet.temperature[0]
        is m.fs.u.control_volume.properties_out[0].temperature
    )


@pytest.mark.unit
def test_add_outlet_port_CV1D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume1DBlock(
        property_package=m.fs.pp,
        transformation_method="dae.finite_difference",
        transformation_scheme="BACKWARD",
    )

    m.fs.u.control_volume.add_geometry()
    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.outlet is p_obj
    assert len(m.fs.u.outlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.outlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties[0, 1].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.outlet.pressure[0] is m.fs.u.control_volume.properties[0, 1].pressure
    assert (
        m.fs.u.outlet.temperature[0]
        is m.fs.u.control_volume.properties[0, 1].temperature
    )


@pytest.mark.unit
def test_add_outlet_port_CV1D_backward():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume1DBlock(
        property_package=m.fs.pp,
        transformation_method="dae.finite_difference",
        transformation_scheme="BACKWARD",
    )

    m.fs.u.control_volume.add_geometry(flow_direction=FlowDirection.backward)
    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port()

    assert isinstance(p_obj, Port)
    assert m.fs.u.outlet is p_obj
    assert len(m.fs.u.outlet) == 1

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert (
                m.fs.u.outlet.component_flow_phase[0, p, j]
                is m.fs.u.control_volume.properties[0, 0].flow_mol_phase_comp[p, j]
            )
    assert m.fs.u.outlet.pressure[0] is m.fs.u.control_volume.properties[0, 0].pressure
    assert (
        m.fs.u.outlet.temperature[0]
        is m.fs.u.control_volume.properties[0, 0].temperature
    )


@pytest.mark.unit
def test_add_outlet_port_CV0D_no_default_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    with pytest.raises(
        ConfigurationError,
        match=re.escape(
            "fs.u add_outlet_port was called without a block "
            "argument but no default ControlVolume exists "
            "(control_volume). Please provide block to which the "
            "Port should be associated."
        ),
    ):
        m.fs.u.add_outlet_port()


@pytest.mark.unit
def test_add_outlet_port_CV0D_full_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port(name="test_port", block=m.fs.u.cv, doc="Test")

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1

    # Set new outlet conditions to differentiate from inlet
    m.fs.u.cv.properties_out[0].a = 10
    m.fs.u.cv.properties_out[0].b = 20
    m.fs.u.cv.properties_out[0].c = 30

    for p in m.fs.pp.phase_list:
        for j in m.fs.pp.component_list:
            assert m.fs.u.test_port.component_flow_phase[0, p, j].value == (
                m.fs.u.cv.properties_out[0].flow_mol_phase_comp[p, j].value
            )
    assert (
        m.fs.u.test_port.pressure[0].value == m.fs.u.cv.properties_out[0].pressure.value
    )
    assert (
        m.fs.u.test_port.temperature[0].value
        == m.fs.u.cv.properties_out[0].temperature.value
    )


@pytest.mark.unit
def test_add_outlet_port_CV0D_part_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(
        ConfigurationError,
        match="fs.u add_outlet_port was called with a block argument, "
        "but a name argument was not provided. Either both "
        "a name and a block must be provided or neither.",
    ):
        m.fs.u.add_outlet_port(block=m.fs.u.cv)

    with pytest.raises(
        ConfigurationError,
        match="fs.u add_outlet_port was called without a block "
        "argument but a name argument was provided. Either "
        "both a name and a block must be provided or neither.",
    ):
        m.fs.u.add_outlet_port(name="foo")


@pytest.mark.unit
def test_add_outlet_port_CV0D_no_state_blocks():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    with pytest.raises(
        ConfigurationError,
        match="fs.u - control volume does not have expected "
        "names for StateBlocks. Please check that the control "
        "volume was constructed correctly.",
    ):
        m.fs.u.add_outlet_port()


@pytest.mark.unit
def test_fix_unfix_initial_conditions():
    fs = Flowsheet(dynamic=True, time_set=[0, 1, 2], time_units=units.s, concrete=True)
    fs._setup_dynamics()

    fs.b = Unit()

    fs.b.material_accumulation = Var(fs.time, ["a", "b", "c"])
    fs.b.element_accumulation = Var(fs.time, ["a", "b", "c"])
    fs.b.energy_accumulation = Var(fs.time)

    fs.fix_initial_conditions()

    for t in fs.time:
        for j in ["a", "b", "c"]:
            if t == 0:
                assert fs.b.material_accumulation[t, j].fixed
                assert fs.b.element_accumulation[t, j].fixed
                assert fs.b.energy_accumulation[t].fixed
            else:
                assert fs.b.material_accumulation[t, j].fixed is False
                assert fs.b.element_accumulation[t, j].fixed is False
                assert fs.b.energy_accumulation[t].fixed is False

    fs.unfix_initial_conditions()

    for t in fs.time:
        for j in ["a", "b", "c"]:
            assert fs.b.material_accumulation[t, j].fixed is False
            assert fs.b.element_accumulation[t, j].fixed is False
            assert fs.b.energy_accumulation[t].fixed is False


@pytest.mark.unit
def test_get_stream_table_contents_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    m.fs.u.add_inlet_port()
    m.fs.u.add_outlet_port()

    df = m.fs.u._get_stream_table_contents()

    assert df.loc["component_flow_phase ('p1', 'c1')"]["Inlet"] == 2
    assert df.loc["component_flow_phase ('p1', 'c2')"]["Inlet"] == 2
    assert df.loc["component_flow_phase ('p2', 'c1')"]["Inlet"] == 2
    assert df.loc["component_flow_phase ('p2', 'c2')"]["Inlet"] == 2
    assert df.loc["pressure"]["Inlet"] == 1e5
    assert df.loc["temperature"]["Inlet"] == 300

    assert df.loc["component_flow_phase ('p1', 'c1')"]["Outlet"] == 2
    assert df.loc["component_flow_phase ('p1', 'c2')"]["Outlet"] == 2
    assert df.loc["component_flow_phase ('p2', 'c1')"]["Outlet"] == 2
    assert df.loc["component_flow_phase ('p2', 'c2')"]["Outlet"] == 2
    assert df.loc["pressure"]["Outlet"] == 1e5
    assert df.loc["temperature"]["Outlet"] == 300


@pytest.mark.unit
def test_get_stream_table_contents_CV0D_missing_default_port():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(property_package=m.fs.pp)

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.u._get_stream_table_contents()


@pytest.mark.unit
def test_add_state_material_balances_invalid_state_1():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb = object()

    with pytest.raises(
        ConfigurationError,
        match="fs.u state_1 argument to add_state_material_balances was "
        "not an instance of a State Block.",
    ):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            state_1=m.fs.u.sb,
            state_2=m.fs.u.sb,
        )


@pytest.mark.unit
def test_add_state_material_balances_invalid_state_2():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = object()

    with pytest.raises(
        ConfigurationError,
        match="fs.u state_2 argument to add_state_material_balances was "
        "not an instance of a State Block.",
    ):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            state_1=m.fs.u.sb1,
            state_2=m.fs.u.sb2,
        )


@pytest.mark.unit
def test_add_state_material_balances_mixed_states():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp1 = PhysicalParameterTestBlock()
    m.fs.pp2 = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp1.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp2.build_state_block(m.fs.time)

    with pytest.raises(
        ConfigurationError,
        match="fs.u add_state_material_balances method was provided with "
        "State Blocks are not linked to the same "
        "instance of a Physical Parameter Block. This method "
        "only supports linking State Blocks from the same "
        "Physical Parameter Block.",
    ):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            state_1=m.fs.u.sb1,
            state_2=m.fs.u.sb2,
        )


@pytest.mark.unit
def test_add_state_material_balances_component_phase():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    m.fs.u.add_state_material_balances(
        balance_type=MaterialBalanceType.componentPhase,
        state_1=m.fs.u.sb1,
        state_2=m.fs.u.sb2,
    )

    assert isinstance(m.fs.u.state_material_balances, Constraint)
    assert len(m.fs.u.state_material_balances) == 4
    for k in m.fs.u.state_material_balances:
        assert k in [
            (0.0, "p1", "c1"),
            (0.0, "p1", "c2"),
            (0.0, "p2", "c1"),
            (0.0, "p2", "c2"),
        ]


@pytest.mark.unit
def test_add_state_material_balances_component_total():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    m.fs.u.add_state_material_balances(
        balance_type=MaterialBalanceType.componentTotal,
        state_1=m.fs.u.sb1,
        state_2=m.fs.u.sb2,
    )

    assert isinstance(m.fs.u.state_material_balances, Constraint)
    assert len(m.fs.u.state_material_balances) == 2
    for k in m.fs.u.state_material_balances:
        assert k in [(0.0, "c1"), (0.0, "c2")]


@pytest.mark.unit
def test_add_state_material_balances_total():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    m.fs.u.add_state_material_balances(
        balance_type=MaterialBalanceType.total, state_1=m.fs.u.sb1, state_2=m.fs.u.sb2
    )

    assert isinstance(m.fs.u.state_material_balances, Constraint)
    assert len(m.fs.u.state_material_balances) == 1


@pytest.mark.unit
def test_add_state_material_balances_element_total():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.elementTotal,
            state_1=m.fs.u.sb1,
            state_2=m.fs.u.sb2,
        )


@pytest.mark.unit
def test_add_state_material_balances_none():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    with pytest.raises(BalanceTypeNotSupportedError):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.none,
            state_1=m.fs.u.sb1,
            state_2=m.fs.u.sb2,
        )


@pytest.mark.unit
def test_add_state_material_balances_double_call():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()

    m.fs.u.sb1 = m.fs.pp.build_state_block(m.fs.time)
    m.fs.u.sb2 = m.fs.pp.build_state_block(m.fs.time)

    m.fs.u.add_state_material_balances(
        balance_type=MaterialBalanceType.componentPhase,
        state_1=m.fs.u.sb1,
        state_2=m.fs.u.sb2,
    )

    with pytest.raises(AttributeError):
        m.fs.u.add_state_material_balances(
            balance_type=MaterialBalanceType.componentPhase,
            state_1=m.fs.u.sb1,
            state_2=m.fs.u.sb2,
        )



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Block-like object meant for controller models.
"""

from idaes.apps.caprese.common.config import (
    ControlPenaltyType,
)
from idaes.apps.caprese.common.config import VariableCategory as VC

from idaes.apps.caprese.nmpc_var import (
    DiffVar,
    AlgVar,
    InputVar,
    DerivVar,
    FixedVar,
    MeasuredVar,
)
from idaes.apps.caprese.dynamic_block import (
    _DynamicBlockData,
    DynamicBlock,
)

from pyomo.environ import (
    Objective,
    check_optimal_termination,
    Constraint,
    Block,
)
from pyomo.common.collections import ComponentMap
from pyomo.dae.set_utils import deactivate_model_at
from pyomo.core.base.indexed_component import UnindexedComponent_set


def pwc_rule(ctrl, i, t):
    time = ctrl.time
    sp_set = set(ctrl.sample_points)
    if t in sp_set:
        # No need to check for time.first() as it is a sample point
        return Constraint.Skip
    t_next = time.next(t)
    return ctrl.vectors.input[i, t_next] == ctrl.vectors.input[i, t]


class _ControllerBlockData(_DynamicBlockData):
    """This class adds methods useful for working with dynamic
    models to be used by a controller. These include methods for
    calculating a setpoint and adding objective functions.
    """

    def solve_setpoint(self, solver, **kwargs):
        """This method performs a "real time optimization-type"
        solve on the model, using only the variables and constraints
        at the first point in time. The purpose is to calculate a
        (possibly steady state) setpoint.

        Parameters:
            solver: A Pyomo solver object that will be used to solve
                    the model with only variables and constraints at t0
                    active.
            require_steady: Whether derivatives should be fixed to zero
                            for the solve. Default is `True`.

        """
        require_steady = kwargs.pop("require_steady", True)
        config = self.CONFIG(kwargs)
        model = self.mod
        time = self.time
        t0 = time.first()

        was_originally_active = ComponentMap(
            [
                (comp, comp.active)
                for comp in model.component_data_objects((Constraint, Block))
            ]
        )
        non_initial_time = list(time)[1:]
        deactivated = deactivate_model_at(
            model,
            time,
            non_initial_time,
            allow_skip=True,
            suppress_warnings=True,
        )
        was_fixed = ComponentMap()

        # Cache "important" values to re-load after solve
        init_input = (
            list(self.vectors.input[:, t0].value) if VC.INPUT in self.categories else []
        )
        init_meas = (
            list(self.vectors.measurement[:, t0].value)
            if VC.MEASUREMENT in self.categories
            else []
        )

        # Fix/unfix variables as appropriate
        # Order matters here. If a derivative is used as an IC, we still want
        # it to be fixed if steady state is required.
        if VC.MEASUREMENT in self.categories:
            self.vectors.measurement[:, t0].unfix()

        if VC.INPUT in self.categories:
            input_vars = self.vectors.input
            was_fixed = ComponentMap((var, var.fixed) for var in input_vars[:, t0])
            input_vars[:, t0].unfix()
        if VC.DERIVATIVE in self.categories:
            if require_steady == True:
                self.vectors.derivative[:, t0].fix(0.0)

        self.setpoint_objective.activate()

        # Solve single-time point optimization problem
        #
        # If these sets do not necessarily exist, then I don't think
        # I can make any assertion about the number of degrees of freedom
        # dof = degrees_of_freedom(model)
        # if require_steady:
        #    assert dof == len(self.INPUT_SET)
        # else:
        #    assert dof == (len(self.INPUT_SET) +
        #            len(self.DIFFERENTIAL_SET))
        results = solver.solve(self, tee=config.tee)
        if check_optimal_termination(results):
            pass
        else:
            msg = "Failed to solve for full state setpoint values"
            raise RuntimeError(msg)

        self.setpoint_objective.deactivate()

        # Revert changes. Again, order matters
        if VC.DERIVATIVE in self.categories:
            if require_steady == True:
                self.vectors.derivative[:, t0].unfix()
        if VC.MEASUREMENT in self.categories:
            self.vectors.measurement[:, t0].fix()

        # Reactivate components that were deactivated
        for t, complist in deactivated.items():
            for comp in complist:
                if was_originally_active[comp]:
                    comp.activate()

        # Fix inputs that were originally fixed
        if VC.INPUT in self.categories:
            for var in self.vectors.input[:, t0]:
                if was_fixed[var]:
                    var.fix()

        setpoint_ctype = (DiffVar, AlgVar, InputVar, FixedVar, DerivVar, MeasuredVar)
        for var in self.component_objects(setpoint_ctype):
            var.setpoint = var[t0].value

        # Restore cached values
        if VC.INPUT in self.categories:
            self.vectors.input.values = init_input
        if VC.MEASUREMENT in self.categories:
            self.vectors.measurement.values = init_meas

    def add_setpoint_objective(
        self,
        setpoint,
        weights,
    ):
        """This method uses user-provided setpoints and weights
        to directly construct a least-squares objective for the
        specified variables at the first time point. The purpose
        is to then use the first time point to solve for a "proper
        setpoint" that provides a value for every variable.

        Parameters:
            setpoint: List of vardata, value tuples describing the
                      setpoint values of these specified variables.
            weights: List of vardata, value tuples describing the
                     weight for each variable's term in the objective.

        """
        vardata_map = self.vardata_map
        for vardata, weight in weights:
            nmpc_var = vardata_map[vardata]
            nmpc_var.weight = weight

        weight_vector = []
        for vardata, sp in setpoint:
            nmpc_var = vardata_map[vardata]
            if nmpc_var.weight is None:
                self.logger.warning("Weight not supplied for %s" % var.name)
                nmpc_var.weight = 1.0
            weight_vector.append(nmpc_var.weight)

        obj_expr = sum(
            weight_vector[i] * (var - sp) ** 2 for i, (var, sp) in enumerate(setpoint)
        )
        self.setpoint_objective = Objective(expr=obj_expr)

    def add_tracking_objective(
        self,
        weights,
        control_penalty_type=ControlPenaltyType.ERROR,
        state_ctypes=DiffVar,
        # TODO: Option for user to provide a setpoint here.
        #       (Should ignore setpoint attrs)
        state_weight=1.0,  # These are liable to get confused with other weights
        input_weight=1.0,
        objective_weight=1.0,
    ):
        """This method constructs a tracking objective based on existing
        `setpoint` attributes of this block's `NmpcVar`s and adds it to the
        block. Only specific ctypes listed in `state_ctypes` are directly
        penalized in the objective. These should be the differential variables
        or the measurements, for traditional NMPC.

        Parameters:
            weights: A list of vardata-value tuples describing the weight to
                     be given to the objective term containing each variable.
            control_penalty_type: An entry of the `ControlPenaltyType` enum,
                                  used to determine if control variable error
                                  or action will be penalized.
            state_ctypes: A ctype for the state variables that should be
                          included in the objective function. This should be
                          one of the types from `nmpc_var.py`.
            state_weight: A scalar weight for the state terms in the objective.
            input_weight: A scalar weight for the input terms in the objective.
            objective_weight: A scalar weight for the entire objective.
        """
        samples = self.sample_points
        # Since t0 ~is~ a sample point, will have to iterate
        # over samples[1:]
        n_sample_points = len(samples)

        # First set the weight of each variable specified
        vardata_map = self.vardata_map
        for vardata, weight in weights:
            var = vardata_map[vardata]
            var.weight = weight

        if not (
            control_penalty_type == ControlPenaltyType.ERROR
            or control_penalty_type == ControlPenaltyType.ACTION
            or control_penalty_type == ControlPenaltyType.NONE
        ):
            raise ValueError(
                "control_penalty_type argument must be 'ACTION', 'ERROR', " "or 'NONE'."
            )

        states = list(self.component_objects(state_ctypes))
        inputs = list(self.component_objects(InputVar))

        state_term = sum(
            state.weight * (state[t] - state.setpoint) ** 2
            for state in states
            if state.weight is not None and state.setpoint is not None
            for t in samples[1:]
        )

        if control_penalty_type == ControlPenaltyType.ERROR:
            input_term = sum(
                var.weight * (var[t] - var.setpoint) ** 2
                for var in inputs
                if var.weight is not None and var.setpoint is not None
                for t in samples[1:]
            )
            obj_expr = objective_weight * (
                state_weight * state_term + input_weight * input_term
            )
        elif control_penalty_type == ControlPenaltyType.ACTION:
            input_term = sum(
                var.weight * (var[samples[k]] - var[samples[k - 1]]) ** 2
                for var in inputs
                if var.weight is not None and var.setpoint is not None
                for k in range(1, n_sample_points)
            )
            obj_expr = objective_weight * (
                state_weight * state_term + input_weight * input_term
            )
        elif control_penalty_type == ControlPenaltyType.NONE:
            obj_expr = objective_weight * state_weight * state_term

        self.tracking_objective = Objective(expr=obj_expr)

    def constrain_control_inputs_piecewise_constant(self):
        """Adds an indexed constraint `pwc_constraint` that
        sets each control variable equal to itself at the previous
        time point. This constraint is skipped at sample points.
        """
        time = self.time
        if VC.INPUT in self.categories:
            input_set = self.INPUT_SET
            self.pwc_constraint = Constraint(input_set, time, rule=pwc_rule)
        else:
            raise RuntimeError(
                "Trying to add constraints on inputs but no input "
                "variables have been specified."
            )


class ControllerBlock(DynamicBlock):
    """This is a user-facing class to be instantiated when one
    wants to work with a block capable of the methods of
    `_ControllerBlockData`.
    """

    _ComponentDataClass = _ControllerBlockData

    def __new__(cls, *args, **kwds):
        # Decide what class to allocate
        if cls != ControllerBlock:
            target_cls = cls
        elif not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            target_cls = SimpleControllerBlock
        else:
            target_cls = IndexedControllerBlock
        return super(ControllerBlock, cls).__new__(target_cls)


class SimpleControllerBlock(_ControllerBlockData, ControllerBlock):
    def __init__(self, *args, **kwds):
        _ControllerBlockData.__init__(self, component=self)
        ControllerBlock.__init__(self, *args, **kwds)

    # Pick up the display() from Block and not BlockData
    display = ControllerBlock.display


class IndexedControllerBlock(ControllerBlock):
    def __init__(self, *args, **kwargs):
        ControllerBlock.__init__(self, *args, **kwargs)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Caprese helper utility functions.
"""
import pytest
from pytest import approx

from pyomo.environ import SolverFactory, Var, value, Reference
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.expr.visitor import identify_variables
from pyomo.dae.flatten import flatten_dae_components

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    activated_equalities_generator,
)
from idaes.apps.caprese.util import *
from idaes.apps.caprese.common.config import NoiseBoundOption
from idaes.apps.caprese.examples.cstr_model import make_model
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

import random

__author__ = "Robert Parker"


# See if ipopt is available and set up solver
solver_available = SolverFactory("ipopt").available()
if solver_available:
    solver = get_solver(
        solver="ipopt",
        options={
            "tol": 1e-6,
            "mu_init": 1e-8,
            "bound_push": 1e-8,
            "halt_on_ampl_error": "yes",
        },
    )
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_by_element_in_range():
    mod = make_model(horizon=2, ntfe=20)
    assert degrees_of_freedom(mod) == 0

    scalar_vars, dae_vars = flatten_dae_components(mod.fs, mod.fs.time, Var)
    diff_vars = [
        Reference(mod.fs.cstr.control_volume.energy_holdup[:, "aq"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "S"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "E"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "C"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "P"]),
    ]

    initialize_by_element_in_range(
        mod.fs,
        mod.fs.time,
        0,
        1,
        solver=solver,
        dae_vars=dae_vars,
        time_linking_variables=diff_vars,
        outlvl=idaeslog.DEBUG,
        solve_initial_conditions=True,
    )

    assert degrees_of_freedom(mod.fs) == 0

    assert mod.fs.cstr.outlet.conc_mol[1, "S"].value == approx(10.189, abs=1e-3)
    assert mod.fs.cstr.outlet.conc_mol[1, "C"].value == approx(0.4275, abs=1e-4)
    assert mod.fs.cstr.outlet.conc_mol[1, "E"].value == approx(0.0541, abs=1e-4)
    assert mod.fs.cstr.outlet.conc_mol[1, "P"].value == approx(0.3503, abs=1e-4)

    initialize_by_element_in_range(
        mod.fs,
        mod.fs.time,
        1,
        2,
        solver=solver,
        dae_vars=dae_vars,
        outlvl=idaeslog.DEBUG,
    )

    assert degrees_of_freedom(mod.fs) == 0
    for con in activated_equalities_generator(mod.fs):
        assert value(con.body) - value(con.upper) < 1e-5

    assert mod.fs.cstr.outlet.conc_mol[2, "S"].value == approx(11.263, abs=1e-3)
    assert mod.fs.cstr.outlet.conc_mol[2, "C"].value == approx(0.4809, abs=1e-4)
    assert mod.fs.cstr.outlet.conc_mol[2, "E"].value == approx(0.0538, abs=1e-4)
    assert mod.fs.cstr.outlet.conc_mol[2, "P"].value == approx(0.4372, abs=1e-4)


@pytest.mark.unit
def test_get_violated_bounds():
    bounds = (1.0, 2.0)
    val = 1.5
    assert get_violated_bounds(val, bounds) == (None, 0)

    val = 0.8
    assert get_violated_bounds(val, bounds) == (1.0, 1)

    val = 2.5
    assert get_violated_bounds(val, bounds) == (2.0, -1)


@pytest.mark.unit
def test_apply_noise():
    random.seed(1234)
    noise_function = random.gauss
    val_list = [1.0, 2.0, 3.0]
    params_list = [0.05, 0.05, 0.05]
    new_vals = apply_noise(val_list, params_list, noise_function)

    for val, new_val, p in zip(val_list, new_vals, params_list):
        assert abs(val - new_val) < 3 * p

    noise_function = lambda val, rad: random.uniform(val - rad, val + rad)
    new_vals = apply_noise(val_list, params_list, noise_function)
    for val, new_val, p in zip(val_list, new_vals, params_list):
        assert abs(val - new_val) <= p


@pytest.mark.unit
def test_apply_bounded_noise_discard():
    NBO = NoiseBoundOption
    random.seed(2345)
    noise_function = random.gauss
    vals = [5.0, 10.0]
    params = [1.0, 2.0]
    bounds_list = [(4.0, 6.0), (9.0, 11.0)]

    N = 100
    results = [set(), set()]
    for _ in range(N):
        # Very low probability we ever need to discard more than 15 times
        newvals = apply_noise_with_bounds(
            vals,
            params,
            noise_function,
            bounds_list,
            bound_option=NBO.DISCARD,
            max_number_discards=15,
        )
        for r, n in zip(results, newvals):
            r.add(n)

    for val, res, (lb, ub) in zip(vals, results, bounds_list):
        # Vanishingly small probability we get the same value twice.
        assert len(res) == 100

        inner_lb = val + (lb - val) / 2
        inner_ub = val + (ub - val) / 2
        n_inner = 0
        n_outer = 0
        for r in res:
            # Vanishingly small probability we land exactly on a bound
            assert lb < r and r < ub
            if inner_lb < val and val < inner_ub:
                n_inner += 1
            else:
                n_outer += 1

        # Expect values to be clustered around the nominal.
        assert n_inner > n_outer

    with pytest.raises(MaxDiscardError):
        for _ in range(N):
            # Very likely that we will eventually need to discard a value
            newvals = apply_noise_with_bounds(
                vals,
                params,
                noise_function,
                bounds_list,
                bound_option=NBO.DISCARD,
                max_number_discards=0,
            )

    with pytest.raises(MaxDiscardError):
        for _ in range(N):
            # In fact, very likely that we will need to discard more than
            # 5 times consecutively over the course of generating 100 numbers
            newvals = apply_noise_with_bounds(
                vals,
                params,
                noise_function,
                bounds_list,
                bound_option=NBO.DISCARD,
                max_number_discards=5,
            )
            # We could calculate the expected max number of discards to generate
            # 100 values with the above mu, sigma, and bounds, but that seems
            # unnecessary. For now I am satisfied that this number seems to be
            # between 5 and 15 for gauss(10., 2.) \in (9., 11.)


@pytest.mark.unit
def test_apply_bounded_noise_push_zero_eps():
    NBO = NoiseBoundOption
    random.seed(3456)
    noise_function = random.gauss
    vals = [5.0, 10.0]
    params = [1.0, 2.0]
    bounds_list = [(4.0, 6.0), (9.0, 11.0)]


@pytest.mark.unit
def test_apply_bounded_noise_push_nonzero_eps():
    NBO = NoiseBoundOption
    random.seed(3456)
    noise_function = random.gauss
    vals = [5.0, 10.0]
    params = [1.0, 2.0]
    bounds_list = [(4.0, 6.0), (9.0, 11.0)]

    N = 100
    results = [[], []]
    for _ in range(N):
        newvals = apply_noise_with_bounds(
            vals,
            params,
            noise_function,
            bounds_list,
            bound_option=NBO.PUSH,
            bound_push=0.0,
        )
        for r, n in zip(results, newvals):
            r.append(n)

    # These flags will be used to make sure we cover both branches of an
    # "if tree" below.
    b1, b2 = False, False

    for val, res, p, (lb, ub) in zip(vals, results, params, bounds_list):
        # Very low probability we DON'T get the same value at least
        # twice. (I.e. we never exceed the same bound more than once.)
        assert len(set(res)) < 100

        n_ub = 0
        n_lb = 0
        n_interior = 0
        for r in res:
            assert lb <= r <= ub
            if r == lb:
                n_lb += 1
            elif r == ub:
                n_ub += 1
            else:
                n_interior += 1

        assert n_lb >= 1
        assert n_ub >= 1
        assert n_interior >= 1

        # Very rough check that the distribution looks something like we
        # expect.
        if (val - lb) >= p and (ub - val) >= p:
            # If our bounds are at least sigma from the mean, we expect more
            # "interior" points than points at the bounds.
            assert n_interior > n_lb + n_ub
            b1 = True
        elif (val - lb) <= p / 2 and (ub - val) <= p / 2:
            # If our bounds are within sigma/2 of the mean, we expect more
            # points at the bounds than in the interior.
            assert n_lb + n_ub > n_interior
            b2 = True
        else:
            raise RuntimeError()

    # Sanity. Make sure we covered both branches.
    assert b1 and b2

    # Now test with a nonzero bound push
    eps_b = 0.01
    results = [[], []]
    for _ in range(N):
        newvals = apply_noise_with_bounds(
            vals,
            params,
            noise_function,
            bounds_list,
            bound_option=NBO.PUSH,
            bound_push=eps_b,
        )
        for r, n in zip(results, newvals):
            r.append(n)

    for val, res, p, (lb, ub) in zip(vals, results, params, bounds_list):
        # Very low probability we DON'T get the same value at least
        # twice. (I.e. we never exceed the same bound more than once.)
        assert len(set(res)) < 100

        n_ub = 0
        n_lb = 0
        for r in res:
            # Satisfy bounds strictly
            assert lb < r < ub
            if r == lb + eps_b:
                n_lb += 1
            elif r == ub - eps_b:
                n_ub += 1

        assert n_lb >= 1
        assert n_ub >= 1


@pytest.mark.unit
def test_apply_bounded_noise_fail():
    NBO = NoiseBoundOption
    random.seed(13456)
    noise_function = random.gauss
    vals = [5.0, 10.0]
    params = [1.0, 1.0]
    bounds_list = [(4.0, 6.0), (9.0, 11.0)]

    N = 10
    results = [[], []]
    with pytest.raises(RuntimeError):
        # This is a very frail option
        for _ in range(N):
            newvals = apply_noise_with_bounds(
                vals, params, noise_function, bounds_list, bound_option=NBO.FAIL
            )
            for r, n in zip(results, newvals):
                r.append(n)

    N = 2
    results = [[], []]
    for _ in range(N):
        newvals = apply_noise_with_bounds(
            vals, params, noise_function, bounds_list, bound_option=NBO.FAIL
        )
        for r, n in zip(results, newvals):
            r.append(n)

    for val, res, p, (lb, ub) in zip(vals, results, params, bounds_list):
        # Expect unique values that don't violate the bounds
        assert len(set(res)) == N
        for r in res:
            assert lb < r < ub


@pytest.mark.component
def _test_add_noise_at_time():
    mod = make_model(horizon=2, ntfe=20)
    time = mod.fs.time
    t0 = time.first()
    assert degrees_of_freedom(mod) == 0

    scalar_vars, dae_vars = flatten_dae_components(mod.fs, time, Var)
    diff_vars = [
        Reference(mod.fs.cstr.control_volume.energy_holdup[:, "aq"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "S"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "E"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "C"]),
        Reference(mod.fs.cstr.control_volume.material_holdup[:, "aq", "P"]),
    ]

    for t in time:
        diff_vars[0][t].setlb(290)
        diff_vars[0][t].setub(310)
        for i in range(1, 5):
            diff_vars[i][t].setlb(0)
            diff_vars[i][t].setub(1)
            # Pretend this is mole fraction...

    assert diff_vars[0][0].value == 300
    for i in range(1, 5):
        assert diff_vars[i][0].value == 0.001

    copy_values_at_time(diff_vars, diff_vars, [t for t in time if t != t0], t0)

    for seed in [4, 8, 15, 16, 23, 42]:
        random.seed(seed)
        weights = [10, 0.001, 0.001, 0.001, 0.001]
        nom_vals = add_noise_at_time(diff_vars, 0, weights=weights)

        assert nom_vals[0][0] == 300
        assert diff_vars[0][0].value != 300
        assert diff_vars[0][0].value == approx(300, abs=2)
        for i in range(1, 5):
            assert nom_vals[0][i] == 0.001
            # ^ nom_vals indexed by time, then var-index. This is confusing,
            # might need to change (or only accept one time point at a time)
            assert diff_vars[i][0].value != 0.001
            assert diff_vars[i][0].value == approx(0.001, abs=2e-4)
            # Within four standard deviations should be a safe check

        for i in range(0, 5):
            diff_vars[i][0].set_value(nom_vals[0][i])
        # Reset and try again with new seed

    # Try providing function for uniform random
    rand_fcn = random.uniform

    random_arg_dict = {
        "range_list": [
            (295, 305),
            (0.001, 0.01),
            (0.001, 0.01),
            (0.001, 0.01),
            (0.001, 0.01),
        ]
    }

    def args_fcn(i, val, **kwargs):
        # args_fcn expects arguments like this
        range_list = kwargs.pop("range_list", None)
        return range_list[i]

    nom_vals = add_noise_at_time(
        diff_vars,
        0.5,
        random_function=rand_fcn,
        args_function=args_fcn,
        random_arg_dict=random_arg_dict,
    )

    assert nom_vals[0.5][0] == 300
    assert diff_vars[0][0.5].value != 300
    assert 295 <= diff_vars[0][0.5].value <= 305
    for i in range(1, 5):
        assert nom_vals[0.5][i] == 0.001
        assert diff_vars[i][0.5].value != 0.001
        assert 0.001 <= diff_vars[i][0.5].value <= 0.01

    # Try to get some bound violations
    random_arg_dict = {"range_list": [(295, 305), (1, 2), (1, 2), (1, 2), (1, 2)]}

    nom_vals = add_noise_at_time(
        diff_vars,
        1,
        random_function=rand_fcn,
        args_function=args_fcn,
        random_arg_dict=random_arg_dict,
        bound_strategy="push",
        bound_push=0.01,
    )

    for i in range(1, 5):
        assert diff_vars[i][1].value == 0.99

    random.seed(123)
    with pytest.raises(ValueError) as exc_test:
        # Large weights - one of these lower bounds should fail...
        nom_vals = add_noise_at_time(
            diff_vars,
            1.5,
            bound_strategy="discard",
            discard_limit=0,
            weights=[1, 1, 1, 1, 1],
            sig_0=0.05,
        )


if __name__ == "__main__":
    test_apply_bounded_noise_push_zero_eps()
    test_apply_bounded_noise_push_nonzero_eps()
    test_apply_bounded_noise_discard()
    test_apply_bounded_noise_fail()
    print("Tests passed!")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
These classes handle the metadata aspects of classes representing
property packages.

Implementors of property packages need to do the following:

1. Create a new class that inherits from
   :class:`idaes.core.property_base.PhysicalParameterBlock`, which in turn
   inherits from :class:`HasPropertyClassMetadata`, in this module.

2. In that class, implement the `define_metadata()` method, inherited from
   :class:`HasPropertyClassMetadata`. This method is called
   automatically, once, when the `get_metadata()` method is first invoked.
   An empty metadata object (an instance of :class:`PropertyClassMetadata`)
   will be passed in, which the method should populate with information about
   properties and default units.

Example::

    from idaes.core.property_base import PhysicalParameterBlock

    class MyPropParams(PhysicalParameterBlock):

        @classmethod
        def define_metadata(cls, meta):
            meta.add_default_units({foo.U.TIME: 'fortnights',
                                   foo.U.MASS: 'stones'})
            meta.add_properties({'under_sea': {'units': 'leagues'},
                                'tentacle_size': {'units': 'yards'}})
            meta.add_required_properties({'under_sea': 'leagues',
                                'tentacle_size': 'yards'})

        # Also, of course, implement the non-metadata methods that
        # do the work of the class.

"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import units
from pyomo.core.base.units_container import _PyomoUnit, InconsistentUnitsError
from pyomo.common.deprecation import deprecation_warning

from idaes.core.util.exceptions import PropertyPackageError
from idaes.core.base.property_set import StandardPropertySet, PropertySetBase
import idaes.logger as idaeslog

__author__ = "Dan Gunter <dkgunter@lbl.gov>, Andrew Lee"

_log = idaeslog.getLogger(__name__)


class HasPropertyClassMetadata(object):
    """Interface for classes that have PropertyClassMetadata."""

    _metadata = None

    @classmethod
    def get_metadata(cls):
        """Get property parameter metadata.

        If the metadata is not defined, this will instantiate a new
        metadata object and call `define_metadata()` to set it up.

        If the metadata is already defined, it will be simply returned.

        Returns:
            PropertyClassMetadata: The metadata
        """
        if cls._metadata is None:
            pcm = PropertyClassMetadata()
            cls.define_metadata(pcm)
            cls._metadata = pcm

            # Check that the metadata was actually populated
            # Check requires looking at private attributes
            # pylint: disable-next=protected-access
            if pcm._properties is None or pcm._default_units is None:
                raise PropertyPackageError(
                    "Property package did not populate all expected metadata."
                )
        return cls._metadata

    @classmethod
    def define_metadata(cls, pcm):
        """Set all the metadata for properties and units.

        This method should be implemented by subclasses.
        In the implementation, they should set information into the
        object provided as an argument.

        Args:
            pcm (PropertyClassMetadata): Add metadata to this object.

        Returns:
            None
        """
        raise NotImplementedError()


class UnitSet(object):
    """
    Object defining the set of recognised quantities in IDAES and their base units.

    Units of measurement are defined by setting units for the seven base SI quantities
    (amount, current, length, luminous intensity, mass, temperature and time), from which units
    for all other quantities are derived. The units of the seven base quantities must be provided
    when instantiating the UnitSet, otherwise base SI units are assumed.

    Units can be accesses by via either a property on the UnitSet (e.g., UnitSet.TIME) or
    via an index on the UnitSet (e.g., UnitSet["time"]).
    """

    _base_quantities = {
        "AMOUNT": units.mol,
        "CURRENT": units.ampere,
        "LENGTH": units.meter,
        "LUMINOUS_INTENSITY": units.candela,
        "MASS": units.kilogram,
        "TEMPERATURE": units.kelvin,
        "TIME": units.seconds,
    }

    def __init__(self):
        self._time = units.seconds
        self._length = units.meter
        self._mass = units.kilogram
        self._amount = units.mole
        self._temperature = units.kelvin
        self._current = units.ampere
        self._luminous_intensity = units.candela

    def set_units(
        self,
        amount: _PyomoUnit = units.mol,
        current: _PyomoUnit = units.ampere,
        length: _PyomoUnit = units.meter,
        luminous_intensity: _PyomoUnit = units.candela,
        mass: _PyomoUnit = units.kilogram,
        temperature: _PyomoUnit = units.kelvin,
        time: _PyomoUnit = units.seconds,
    ):
        """
        Set desired units of measurement for the seven base quantities.

        Args:
            amount: units for amount (default = moles)
            current: units for current (default = Amperes)
            length: units for length (default = meters)
            luminous_intensity: units for luminous intensity (default = candela)
            mass: units for mass (default = kilograms)
            temperature: units for temperature (default = Kelvins)
            time: units for time (default = seconds)

        Returns:
            None
        """
        self._time = time
        self._length = length
        self._mass = mass
        self._amount = amount
        self._temperature = temperature
        self._current = current
        self._luminous_intensity = luminous_intensity

        # Check that valid units were assigned
        for q, expected_dim in self._base_quantities.items():
            u = getattr(self, q)
            if not isinstance(u, _PyomoUnit):
                # Check for non-unit inputs from user
                raise PropertyPackageError(
                    f"Unrecognized units of measurement for quantity {q} ({u})"
                )

            # Check for expected dimensionality
            try:
                # Try to convert user-input to SI units of expected dimensions
                units.convert(u, expected_dim)
            except InconsistentUnitsError:
                # An error indicates a mismatch in units or the units registry
                raise PropertyPackageError(
                    f"Invalid units of measurement for quantity {q} ({u}). "
                    "Please ensure units provided are valid for this quantity and "
                    "use the Pyomo unit registry."
                )

    def __getitem__(self, key: str):
        try:
            # Check to catch cases where luminous intensity has a space
            return getattr(self, key.upper().replace(" ", "_"))
        except AttributeError:
            raise PropertyPackageError(
                f"Unrecognised quantity {key}. Please check that this is a recognised quantity "
                "defined in idaes.core.base.property_meta.UnitSet."
            )

    def unitset_is_consistent(self, other: "UnitSet"):
        """
        Checks that defined units of measurement for base quantities are consistent with those
        in other UnitSet.

        Args:
            other: UnitSet to check for consistency with

        Returns:
            Bool indicating whether units are consistent
        """
        return all(getattr(self, q) is getattr(other, q) for q in self._base_quantities)

    @property
    def TIME(self):
        return self._time

    @property
    def LENGTH(self):
        return self._length

    @property
    def MASS(self):
        return self._mass

    @property
    def AMOUNT(self):
        return self._amount

    @property
    def TEMPERATURE(self):
        return self._temperature

    @property
    def CURRENT(self):
        return self._current

    @property
    def LUMINOUS_INTENSITY(self):
        return self._luminous_intensity

    # Length based
    @property
    def AREA(self):
        return self._length**2

    @property
    def VOLUME(self):
        return self._length**3

    @property
    def VOLUME_MASS(self):
        return self._length**3 * self._mass**-1

    @property
    def VOLUME_MOLE(self):
        return self._length**3 * self._amount**-1

    # Backward compatibility name
    @property
    def MOLAR_VOLUME(self):
        msg = "The unit name MOLAR_VOLUME is being deprecated in favor of VOLUME_MOL."
        deprecation_warning(msg=msg, logger=_log, version="2.3.0", remove_in="3.0.0")
        return self.VOLUME_MOLE

    # Flows
    @property
    def FLOW_MASS(self):
        return self._mass * self._time**-1

    @property
    def FLOW_MOLE(self):
        return self._amount * self._time**-1

    @property
    def FLOW_VOL(self):
        return self._length**3 * self._time**-1

    @property
    def FLUX_MASS(self):
        return self._mass * self._time**-1 * self._length**-2

    @property
    def FLUX_MOLE(self):
        return self._amount * self._time**-1 * self._length**-2

    @property
    def FLUX_ENERGY(self):
        return self._mass * self._time**-3

    # Velocity, Acceleration and Force
    @property
    def VELOCITY(self):
        return self._length * self._time**-1

    @property
    def ACCELERATION(self):
        return self._length * self._time**-2

    @property
    def FORCE(self):
        return self._length * self._mass * self._time**-2

    # Pressures
    @property
    def PRESSURE(self):
        return self._mass * self._length**-1 * self._time**-2

    @property
    def GAS_CONSTANT(self):
        return (
            self._mass
            * self._length**2
            * self._time**-2
            * self._temperature**-1
            * self._amount**-1
        )

    # Densities & Concentrations
    @property
    def DENSITY_MASS(self):
        return self._mass * self._length**-3

    @property
    def DENSITY_MOLE(self):
        return self._amount * self._length**-3

    @property
    def MOLALITY(self):
        return self._amount * self._mass

    @property
    def MOLECULAR_WEIGHT(self):
        return self._mass / self._amount

    # Energy
    @property
    def ENERGY(self):
        return self._mass * self._length**2 * self._time**-2

    @property
    def ENERGY_MASS(self):
        return self._length**2 * self._time**-2

    @property
    def ENERGY_MOLE(self):
        return self._mass * self._length**2 * self._time**-2 * self._amount**-1

    @property
    def POWER(self):
        return self._mass * self._length**2 * self._time**-3

    @property
    def VOLTAGE(self):
        return self._mass * self._length**2 * self._time**-3 * self._current**-1

    # Heat Related
    @property
    def HEAT_CAPACITY_MASS(self):
        return self._length**2 * self._time**-2 * self._temperature**-1

    @property
    def HEAT_CAPACITY_MOLE(self):
        return (
            self._mass
            * self._length**2
            * self._time**-2
            * self._temperature**-1
            * self._amount**-1
        )

    @property
    def HEAT_TRANSFER_COEFFICIENT(self):
        return self._mass * self._time**-3 * self._temperature**-1

    # Entropy
    @property
    def ENTROPY(self):
        return self._mass * self._length**2 * self._time**-2 * self._temperature**-1

    @property
    def ENTROPY_MASS(self):
        return self._length**2 * self._time**-2 * self._temperature**-1

    @property
    def ENTROPY_MOLE(self):
        return (
            self._mass
            * self._length**2
            * self._time**-2
            * self._temperature**-1
            * self._amount**-1
        )

    # Transport Properties
    @property
    def DIFFUSIVITY(self):
        return self._length**2 * self._time**-1

    @property
    def DYNAMIC_VISCOSITY(self):
        return self._mass * self._length**-1 * self._time**-1

    @property
    def KINEMATIC_VISCOSITY(self):
        return self._length**2 * self._time**-1

    @property
    def SURFACE_TENSION(self):
        return self._mass * self._time**-2

    @property
    def THERMAL_CONDUCTIVITY(self):
        return self._mass * self._length * self._time**-3 * self._temperature**-1


class PropertyClassMetadata(object):
    """
    Container for metadata about the property class, which includes
    default units and properties.

    Example usage::

        foo = PropertyClassMetadata()
        foo.add_default_units(time = pyo.units.fortnights,
                              mass = pyo.units.stones)
        foo.add_properties({'under_sea': {'method': 'submarine', 'units': 'leagues', 'required': False, 'supported': True},
                            'tentacle_size': {'method': 'kraken', 'units': 'yards', 'required': True, 'supported': True}})

    """

    def __init__(self):
        # TODO: Deprecate in favour of common units property
        self._default_units = UnitSet()
        # Assume a default PropertySet to begin with. Property packages can replace this
        # with more specialized forms if required
        self._properties = StandardPropertySet(parent=self)

    def define_property_set(self, propset: PropertySetBase):
        """
        Define the type of property set to use for this package.

        Args:
            propset: PropertySet class (must derive from PropertySetBase)

        Returns:
            None
        """
        if not issubclass(propset, PropertySetBase):
            raise PropertyPackageError(
                f"{propset} does not derive from IDAES PropertySetBase class."
            )
        self._properties = propset(parent=self)

    @property
    def default_units(self):
        # TODO: Deprecate in favour of common units property
        return self._default_units

    @property
    def derived_units(self):
        # TODO: Deprecate in favour of common units property
        return self._default_units

    @property
    def properties(self):
        return self._properties

    def add_default_units(self, u: dict):
        """
        Set units of measurement for base quantities used in this property package. Units
        should be provided as a dict with keys being the seven base quantities and values
        being Pyomo unit expressions. These will be used to update the UnitSet associated
        with this property package.

        Args:
            u (dict): Key=property, Value=units

        Returns:
            None

        Raises:
            TypeError if definitions for unexpected quantities are found
        """
        # TODO: Could look at replacing dict with defined arguments
        # This would be a big API change
        try:
            self._default_units.set_units(**u)
        except TypeError:
            raise TypeError(
                "Unexpected argument for base quantities found when creating UnitSet. "
                "Please ensure that units are only defined for the seven base quantities."
            )

    def add_properties(self, p: dict):
        """Add properties to the metadata.

        For each property, the value should be another dict which may contain
        the following keys:

        - 'units': (optional) units of measurement for the property.
        - 'indices': (optional) list of sub-property indices for this property. If None, use default set, if False unindexed.
        - 'method': (optional, only if 'indices' is None or False) the name of a method to construct the
                    property as a str, or None if the property will be
                    constructed by default.
        - 'supported': (optional, only if 'indices' is None or False) bool indicating if this property is
                       supported by this package.
        - 'required': (optional, only if 'indices' is None or False) bool indicating if this property is
                      required by this package.
        - 'valid_range': (optional, only if 'indices' is None or False) 2-tuple containing range of validity for
                      property values (lower, upper).
        - 'initialize': (optional) dict indicating 'method', 'required', 'supported' and 'valid_range' values for sub-properties by index.

        Args:
            p (dict): Key=property, Value=dict

        Returns:
            None
        """
        # TODO: Deprecate in favour of directly updating or adding metadata
        for k, v in p.items():
            units = v.pop("units", None)
            try:
                try:
                    n, i = self._properties.get_name_and_index(k)
                except ValueError:
                    msg = (
                        f"The property name {k} in property metadata is not a recognized "
                        "standard property name defined in this PropertySet. Please refer "
                        "to IDAES standard names in the IDAES documentation. You can use "
                        "the define_custom_properties() rather than the add_properties() "
                        "method to define metadata for this property. You can also use a "
                        "different property set by calling the define_property_set() method."
                    )
                    deprecation_warning(
                        msg=msg, logger=_log, version="2.0.0", remove_in="3.0.0"
                    )
                    n = k
                    i = None
                getattr(self._properties, n)[i].update_property(**v)
            except AttributeError:
                # TODO: Deprecate this and make it raise an exception if an unknown property is encountered
                # Force users to explicitly declare new/custom properties
                self._properties.define_property(name=k, **v, units=units)

    def define_custom_properties(self, p: dict):
        """Add custom properties to the metadata.

        For each property, the value should be another dict which may contain
        the following keys:

        - 'units': (optional) units of measurement for the property.
        - 'indices': (optional) list of sub-property indices for this property. If None, use default set, if False unindexed.
        - 'method': (optional, only if 'indices' is None or False) the name of a method to construct the
                    property as a str, or None if the property will be
                    constructed by default.
        - 'supported': (optional, only if 'indices' is None or False) bool indicating if this property is
                       supported by this package.
        - 'required': (optional, only if 'indices' is None or False bool indicating if this property is
                      required by this package.
        - 'initialize': (optional) dict indicating 'method', 'required' and 'supported' values for sub-properties by index.

        Args:
            p (dict): Key=property, Value=dict

        Returns:
            None
        """
        for k, v in p.items():
            self._properties.define_property(name=k, **v)

    def add_required_properties(self, p: str):
        # TODO: Deprecate
        """Add required properties to the metadata.

        Update 'required' attribute of specified properties.
        Note that argument must be a dict for backwards compatibility.

        Args:
            p (dict): Key=property, Value=(ignored)

        Returns:
            None
        """
        for k in p.keys():
            try:
                self._properties[k].set_required(True)
            except KeyError:
                self._properties.define_property(name=k, supported=False, required=True)

    def get_derived_units(self, units: str):
        # TODO: Deprecate in favour of common units property
        return self.derived_units[units]
